/*      
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/bh_correction.cpp $
 $LastChangedRevision: 118 $
 $Date: 2014-07-09 14:22:29 -0400 (Wed, 09 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 * Description:
 *
 * The class bh_correction is responsible for performing beam hardening
 * correction for polychromatic input data. Given the prjfil that was calculated
 * based on weighted average of attenuations at different energy levels and 
 * corrections coefficients [1,2] it approximates monochromatic projection data 
 * which is then used by the reconstruction algorithms.
 *
 * 
 *
 * References:
 * [1] G. T. Herman, Image Reconstruction from Projections, 1980
 * [2] snark14 manual
 *
 */

#include <cstdlib>
#include <cstdio>
#include <cmath>

#include <unistd.h>
#include <sys/types.h>

#include <fstream>
#include <string.h>

#include "bh_correction.h"

using namespace std;

int bh_correction::readInputFilePoly()
{

	BOOLEAN eol, eol1, eol2;
	int i, j;   //loop counter variables

	//read in the number of iterations
	noIter = InFile.getint(TRUE, &eol);
	if (eol == TRUE) //too few values provided
	{
		fprintf(output, "\n **** ERROR: number of iterations missing\n"
				" **** program aborted\n\n");
		exit(-1);
	}
	fprintf(output, "\n Number of iterations for beam hardening "
			"correction set to r = %i\n\n", noIter);

	//read in the degree of the polynomial that approximates monochromatic
	//projections given polychromatic projections
	noDegree = InFile.getint(FALSE, &eol);
	if (eol == TRUE) //too few values provided
	{
		fprintf(output, "\n **** ERROR: the degree is missing\n"
				" **** program aborted\n\n");
		exit(-1);
	}
	fprintf(output, "\n The degree of the polynomial for beam hardening "
			"correction set to  %i\n\n", noDegree);

	//read in the  coefficients of polynomial that approximates monochromatic
	//projections given polychromatic projections
	a = new REAL[noDegree + 1];
	a[0] = InFile.getnum(TRUE, &eol);
	if (eol == TRUE) //too few values provided
	{
		fprintf(output, "\n **** ERROR: too few polynomial coefficients \n"
				" **** program aborted\n\n");
		exit(-1);
	}
	for (i = 1; i <= noDegree; i++)
	{
		a[i] = InFile.getnum(FALSE, &eol);
		if (eol == TRUE) //too few values provided
		{
			fprintf(output, "\n **** ERROR: too few polynomial coefficients \n"
					" **** program aborted\n\n");
			exit(-1);
		}
	}

	if (noIter != 0)
	{
		//read in the number of energy levels
		//this should be the same as Spctrm.nergy, otherwise exit
		nerg = InFile.getint(TRUE, &eol);
		if (eol == TRUE) //too few values provided
		{
			fprintf(output, "\n **** ERROR: the nerg is missing\n"
					" **** program aborted\n\n");
			exit(-1);
		}

		if (nerg != Spctrm.nergy) //these two should be the same
		{
			fprintf(output, "\n **** ERROR: the value of nerg does not agree\n"
					"with the nomber of energy levels written on file11\n"
					" **** program aborted\n\n");
			exit(-1);
		}
		fprintf(output, "\n The number of energy levels is %i\n\n", nerg);

		//read in the number of points for piecewise linear conversion of
		//phantom data from energy[0] to other energy levels
		noPointsConv = InFile.getint(TRUE, &eol);
		if (eol == TRUE) //too few values provided
		{
			fprintf(output, "\n **** ERROR: the noPointsConv is missing\n"
					" **** program aborted\n\n");
			exit(-1);
		}
		fprintf(output,
				"\n The number of points for conversion of the polynomial "
						"for beam hardening correction set to %i\n\n",
				noPointsConv);

		//increase noPointsConv by 1 since we need point (0,0)
		noPointsConv++;
		//read in the  points used for constant linear conversion of
		//phantom data from energy[0] to other energy levels
		for (j = 1; j < Spctrm.nergy; j++)
		{
			//skip j=0 since there is no conversion from the energy[0] to energy[0]
			c[j] = new point[noPointsConv];

			c[j][0].x = 0.0; //these are always 0.0
			c[j][0].y = 0.0; //because vacum attenuates nothing at all energy levels

			for (i = 1; i < noPointsConv; i++)
			{
				c[j][i].x = InFile.getnum(TRUE, &eol1);
				c[j][i].y = InFile.getnum(FALSE, &eol2);
				if (eol1 == TRUE || eol2 == TRUE) //too few values provided
				{
					fprintf(output, "\n **** ERROR: too few point for "
							"energy level %i \n **** program aborted\n\n", j);
					exit(-1);
				}
			}
		}
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////

int bh_correction::correction(SnarkPrjFile * BHC_PrjFile)
{
	INTEGER i, j, k; //loop counters
	INTEGER r = 0;
	REAL x, y, u, den; //store elementary object descriptions
	char command[256]; //store command line to be used in system() call
	char filename[256]; //store file name created on the run
	char line[256]; //store temporarly a line of a file

	if (noIter == 0)
	{
		sprintf(command, "cp %s/q_p0 prjfil", tempDir);
		if (system(command) != 0)
		{
			fprintf(output, "\n **** ERROR: could not execute  command"
					"%s\n  **** program aborted\n\n", command);
			exit(-1);
		}
		return 0;
	}

	//an input file for reconstruction part of iteration,
	//uses existing prjfil to generate recfil
	ofstream bhcInputFromPrjfil;
	//an array of input files to be created based on the recfil created by
	//reconstruction step during an iterative run
	//[0] will be used for monochromatic input file
	//[1] will be used for polychromatic input file
	ofstream bhcInputFiles[2];

	//a copy of the original input file for the main snark14 run
	//it is needed here for information about projection generation, etc.
	ifstream inputFile;

	//a recfil from the itermediate reconstruction during the iterative process
	DIGFileSnarkRec bhcRecfil;

	unsigned int NoOfPixels, NoOfRecSets;

	//a reconstruction image for the intermediate reconstruction during
	//the iterative process
	REAL * bhcRecBuffer;

	/* make sure that the nergy prjfil's are present,
	 * this projfil's contain the origila projection data corrected by multi-
	 * plying by the correction coefficient provided by the user in the input
	 * file
	 */

	/* create nergy input files that use the same structures as the original
	 * input file, use PROJECTION OLD and run convolution algorithm for
	 * reconstruction
	 */

	//create and open input files
	//copy the original input file to the temp directory and open it for reading
	sprintf(command, " cp file11 %s/inputFile.in                    ", tempDir);
	if (system(command) != 0)
	{
		fprintf(output,
				"\n **** ERROR: could not copy input file to inputFile.in\n "
						" **** program aborted\n\n");
		exit(-1);
	}

	//open a new input file for writing

	sprintf(filename, "%s/bhc_in.in", tempDir);
	bhcInputFromPrjfil.open(filename);
	if (!bhcInputFromPrjfil.is_open())
	{
		fprintf(output, "\n **** ERROR: could not open file %s from "
				"temp directory\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	//create input files that use the existing prjfil's to perform the reconstruction
	bhcInputFromPrjfil << "PICTURE RECONSTUCTION " << GeoPar.nelem << " "
			<< GeoPar.pixsiz << "\n";
	bhcInputFromPrjfil << "PROJECTION OLD \n";
	bhcInputFromPrjfil << "EXECUTE CONV \n";
	bhcInputFromPrjfil << "monochromatic intermediate reconstruction\n";
	bhcInputFromPrjfil << "hamming 0.54 2 \n";
	bhcInputFromPrjfil << "END\n";

	//close the input file
	bhcInputFromPrjfil.close();

	//allocalte needed memory
	//bhcRecfils = new DIGFileSnarkRec[Spctrm.nergy];
	//bhcRec_buffer = new REAL * [Spctrm.nergy];
	NoOfPixels = GeoPar.nelem * GeoPar.nelem;

	/* repeat noIter times
	 */
	for (r = 0; r < noIter; r++)
	{
		sprintf(filename, "%s/inputFile.in", tempDir);
		inputFile.open(filename);
		if (!inputFile.is_open())
		{
			fprintf(output,
					"\n **** ERROR: could not read the inputFile.in from "
							"temp directory\n  **** program aborted\n\n");
			exit(-1);
		}

		/* run snark14 and create new input files based on reconstruction images,
		 * then create monochromatic projections (nergy of them) and polychromatic
		 * projections in order to perform the next step of iterative data refinement
		 *
		 * correct the data in projfil's and prepare the input files for the next
		 * round of iterative data refinement step
		 */

		//run snark14 to obtain reconstruction in recfil
		//for (i = 0; i < Spctrm.nergy; i++) {
		//rename the existing q_p%i to prjfil
		sprintf(command, "cd %s\n cp q_p%i  prjfil", tempDir, r);
		if (system(command) != 0)
		{
			fprintf(output, "\n **** ERROR: could not execute  command "
					"%s\n  **** program aborted\n\n", command);
			exit(-1);
		}
		//run snark14 with input file in"i".in
		sprintf(command, "cd %s\n snark14 bhc_in.in >out.out", tempDir);
		if (system(command) != 0)
		{
			fprintf(output, "\n **** ERROR: could not execute  command "
					"%s\n  **** program aborted\n\n", command);
			exit(-1);
		}
		//rename the recfil, so it is not lost at the next run of snark14
		sprintf(command, "cd %s\n mv recfil recfil%i", tempDir, r);
		if (system(command) != 0)
		{
			fprintf(output, "\n **** ERROR: could not execute  command "
					"%s\n  **** program aborted\n\n", command);
			exit(-1);
		}
		//}

		//open just created recfil's and prepare data for reading

		sprintf(filename, "%s/recfil%i", tempDir, r);
		if (bhcRecfil.Open(filename) != 0)
		{
			fprintf(output, "\n **** ERROR: could not open file "
					"%s (r = %i)\n  **** program aborted\n\n", filename, r);
			exit(-1);
		}

		// verify that there is only one reconstruction set (convolution
		// produces only one)

		if (bhcRecfil.GetNoOfRecSets(&NoOfRecSets) != 0)
		{
			fprintf(output,
					"\n **** ERROR: could not read the number of reconstruction"
							" sets from file %s\n  **** program aborted\n\n",
					filename);
			exit(-1);
		}
		if (NoOfRecSets != 1)
		{
			fprintf(output, "\n **** ERROR: the number (= %i) of reconstruction"
					" sets from file %s is incorrect, should be =1\n  "
					"**** program aborted\n\n", NoOfRecSets, filename);
			exit(-1);
		}

		//allocate space for reconstuction data
		bhcRecBuffer = new REAL[NoOfPixels];

		//select the reconstruction
		if (bhcRecfil.SelectRec(0) != 0)
		{
			fprintf(output, "\n **** ERROR: could not select reconstruction #0 "
					" from file %s "
					"**** program aborted\n\n", filename);
			exit(-1);
		}

		//read in the data from the reconstruction pixel by pixel
		bhcRecfil.GetRecData(bhcRecBuffer);

		//close the recfil
		bhcRecfil.Close();

		//}

		//  create new set of input files for monochromatic projections (overwrite the old ones)
		//with structures being pixel by pixel
		//description based on the reconstruction images from recfil's above;
		//  create an input file for polychromatic projection with density
		//values based on densities obtained in the reconstruction recfil's
		//and distibution based on the original input file

		//open the monochromatic input file for writing
		//and the polychromatic input file for writing
		for (i = 0; i < 2; i++)
		{
			sprintf(filename, "%s/%iin_rec%i.in", tempDir, r, i);
			bhcInputFiles[i].open(filename);
			if (!bhcInputFiles[i].is_open())
			{
				fprintf(output, "\n **** ERROR: could not open file %s from "
						"temp directory\n  **** program aborted\n\n", filename);
				exit(-1);
			}
		}

		//start copying initial lines from the original input file to
		//the new input files
		//  until lines
		//       SPECTRUM POLYCHROMATIC imax
		//       d1 p1 d2 p2  ... d1imax  pimax
		// are encountered
		for (i = 0; i < 2; i++)
		{
			bhcInputFiles[i] << "CREATE\n";
		}

		while (!inputFile.eof())
		{
			inputFile.getline(line, 255);
			if (!strncasecmp(line, "SPEC", 4))
			{        //if line starts with the word "SPECTRUM"
				break; //jump out of the loop
			}
			//otherwise write the line to all the newly created input files
			for (i = 0; i < 2; i++)
			{
				bhcInputFiles[i] << line << "\n";
			}
		}
		//write the line
		//       SPECTRUM MONOCHROMATIC "Spctrm.energy[i]
		// to the monochromatic input files

		bhcInputFiles[0] << "SPECTRUM MONOCHROMATIC   " << Spctrm.energy[0]
				<< "\n";

		//write the lines
		//       SPECTRUM POLYCHROMATIC "Spctrm.nergy"
		//       Spctrm.energy[0]  Spctrm.engwt[0] ... Spctrm.energy[nergy-1]  Spctrm.engwt[nergy-1]
		// to the polychromatic input file
		bhcInputFiles[1] << "SPECTRUM POLYCHROMATIC  " << Spctrm.nergy << "\n";
		for (i = 0; i < Spctrm.nergy; i++)
		{
			bhcInputFiles[1] << Spctrm.energy[i] << "  "
					<< 100 * Spctrm.engwt[i] << "   ";
		}
		bhcInputFiles[1] << "\n";
		//write the line
		//       OBJECTS
		// to the input files
		for (i = 0; i < 2; i++)
		{
			bhcInputFiles[i] << "OBJECTS" << "\n";
		}

		//write the pixel by pixel description of the reconstruction image
		// to the input files
		INTEGER sizeHalf = 0.5 * GeoPar.nelem;

		for (k = 0; k < NoOfPixels; k++)
		{
			x = (k % GeoPar.nelem - sizeHalf) * GeoPar.pixsiz;
			y = (k / GeoPar.nelem - sizeHalf) * GeoPar.pixsiz;
			u = GeoPar.pixsiz / 2;
			den = bhcRecBuffer[k];

			bhcInputFiles[0] << "RECT\t" << x << "\t" << y << "\t" << u << "\t"
					<< u << "\t" << 0 << "\t" << den << "\n";

			bhcInputFiles[1] << "RECT\t" << x << "\t" << y << "\t" << u << "\t"
					<< u << "\t" << 0 << "\t";

			bhcInputFiles[1] << den << "\nDENS";

			//compute the densities using piecewise linear conversion function
			for (i = 1; i < Spctrm.nergy; i++)
			{
				//REAL denden = 1.0;
				REAL nextden = 0.0;
				REAL m, mm;     //slope and y intercept of the line segment

				for (j = 1; j < noPointsConv; j++)
				{
					if (den < c[i][j].x
							|| (j == noPointsConv - 1 && den > c[i][j].x))
					{
						m = (c[i][j].y - c[i][j - 1].y)
								/ (c[i][j].x - c[i][j - 1].x);
						mm = c[i][j].y - m * c[i][j].x;
						nextden = m * den + mm;
						break;
					}
					if (den == c[i][j].x)
					{
						nextden = c[i][j].y;
						break;
					}

				}
				bhcInputFiles[1] << "\t" << nextden;
			}

			bhcInputFiles[1] << "\n";
		}

		//write the lines in input files up to line
		//      PROJECTION PSEUDO
		for (i = 0; i < 2; i++)
		{
			bhcInputFiles[i] << "LAST 1 \n" << "PHANTOM  AVERAGE "
					<< Creacm.nave1 << "\n" << GeoPar.nelem << "    "
					<< GeoPar.pixsiz << "\n" << "RAYSUM \n" << "PICTURE TEST \n"
					<< "PROJECTION BHPSEUDO \n"
					<< "psudo run to create prjfil for correction\n";
		}

		//skip through lines of the original input file until line
		//   GEOMETRY ....
		//is encountered
		while (!inputFile.eof())
		{
			inputFile.getline(line, 256);
			if (!strncasecmp(line, "GEOM", 4))
			{        //if line starts with the word "GEOM"
				break; //jump out of the loop
			}
		}

		//copy the lines from the original input file to all the new
		// input files until the line
		//    BACKGROUND ...
		//is encountered
		for (i = 0; i < 2; i++)
		{
			bhcInputFiles[i] << line << "\n";
		}
		while (!inputFile.eof())
		{
			inputFile.getline(line, 255);
			//write the line to all the newly created input files
			for (i = 0; i < 2; i++)
			{
				bhcInputFiles[i] << line << "\n";
			}
			if (!strncasecmp(line, "BACK", 4))
			{        //if line starts with the word "BACKGROUND"
				break; //jump out of the loop
			}
		}

		//finish the files with END line and close the files
		for (i = 0; i < 2; i++)
		{
			bhcInputFiles[i] << "END \n";
			bhcInputFiles[i].close();
		}
		//close the original input file
		inputFile.close();

		//obtain projection data from the input files created above

		for (i = 0; i < 2; i++)
		{

			//run snark14 with input file %rin"i".in
			sprintf(command, "cd %s\n snark14 %iin_rec%i.in >out%i.out",
					tempDir, r, i, i);
			if (system(command) != 0)
			{
				fprintf(output, "\n **** ERROR: could not execute  command"
						"%s\n  **** program aborted\n\n", command);
				exit(-1);
			}
			//rename the prjfil, so it is not lost at the next run of snark14
			if (i == 1)
				sprintf(command, "cd %s\n mv prjfil p%i", tempDir, r);
			else
				sprintf(command, "cd %s\n mv prjfil m%i", tempDir, r);
			if (system(command) != 0)
			{
				fprintf(output, "\n **** ERROR: could not execute  command"
						"%s\n  **** program aborted\n\n", command);
				exit(-1);
			}
		}

		//correct the projection data following the formula
		//m[i] = miu q[p][i] + (m'[i] - miu q(p')[i])
		idr_correction(r);

	} //end of repeat noIter times

	fprintf(output, "\n **** Finished the iterations\n\n");

	/* write the final value of m[0] to prjfil that will be passed to the remainder
	 * of the run of snark14
	 */

	sprintf(command, "cp %s/q_p%i prjfil", tempDir, noIter);
	if (system(command) != 0)
	{
		fprintf(output, "\n **** ERROR: could not execute  command"
				"%s\n  **** program aborted\n\n", command);
		exit(-1);
	}

	//if (bhcRecfil != NULL) delete [] bhcRecfil;
	//if (bhc_InputFiles != NULL) delete [] bhc_InputFiles;
	if (bhcRecBuffer != NULL)
		delete[] bhcRecBuffer;
	return -1;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

int bh_correction::idr_correction( INTEGER r)
{
	//correct the projection data following the formula
	//m[i] = miu q[p][i] + (m'[i] - miu q(p')[i])
	REAL miu = 1.0; //the relaxation parameter, for now permanently set to 1.0
	unsigned INTEGER i, j, k;  //loop counters
	DIGFileSnarkProj Org, NewM, NewP, Corrected;
	char filename[256];
	//char command[256];
	REAL** Buffer = new REAL*[4]; //buffer to store a projection data

	unsigned int NoOfRays;
	unsigned int NoOfProjections;
	ProjFileMH MainHeader;
	ProjFileAH ApplicationHeader;
	ProjFilePrjH ProjectionHeader[3];

	//open original prjfil corrected by coefficient at energy level 0
	//   bhc/q_p0
	//read Main Header and Application Header of the file
	sprintf(filename, "%s/q_p0", tempDir);
	if (Org.Open(filename) != 0)
	{
		fprintf(output, "\n **** ERROR: could not open file "
				"%s (Org)\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	if (Org.GetNoOfRays(&NoOfRays) != 0)
	{
		fprintf(output,
				"\n **** ERROR: could not read the number of rays from  file "
						"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	if (Org.GetNoOfProjs(&NoOfProjections) != 0)
	{
		fprintf(output,
				"\n **** ERROR: could not read the number of projections from  file "
						"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	if (Org.GetMainHeader(&MainHeader) != 0)
	{
		fprintf(output, "\n **** ERROR: could not read Main Header of  file "
				"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	if (Org.GetAppHeader(&ApplicationHeader) != 0)
	{
		fprintf(output,
				"\n **** ERROR: could not read Application Header of  file "
						"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	//open just created monochromatic prjfil at energy level i
	//   bhc/m%r
	sprintf(filename, "%s/m%i", tempDir, r);
	if (NewM.Open(filename) != 0)
	{
		fprintf(output, "\n **** ERROR: could not open file "
				"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	//open just created polychromatic prjfil
	//   bhc/p%r
	sprintf(filename, "%s/p%i", tempDir, r);
	if (NewP.Open(filename) != 0)
	{
		fprintf(output, "\n **** ERROR: could not open file "
				"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	//open prjfil that will contain corrected data,
	//   bhc/q_p%r
	//set its Main Header and Application Header
	sprintf(filename, "%s/q_p%i", tempDir, r + 1);
	if (Corrected.Open(filename, &MainHeader) != 0)
	{
		fprintf(output, "\n **** ERROR: could not open file "
				"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	if (Corrected.SetAppHeader(&ApplicationHeader) != 0)
	{
		fprintf(output,
				"\n **** ERROR: could not set Application Header of file "
						"%s\n  **** program aborted\n\n", filename);
		exit(-1);
	}

	//allocate memory to store a single projection data
	for (j = 0; j < 4; j++)
	{
		(Buffer[j]) = new REAL[NoOfRays];
	}

	for (j = 0; j < NoOfProjections; j++)
	{

		// array set selection
		if (Org.SelectProj(j) != 0)
		{
			fprintf(output, "\n **** ERROR: could not select projection of file"
					"Org\n  **** program aborted\n\n");
			exit(-1);
		}
		//get projection header
		if (Org.GetProjHeader(&ProjectionHeader[0]) != 0)
		{
			fprintf(output,
					"\n **** ERROR: could not get projection header from file"
							"Org\n  **** program aborted\n\n");
			exit(-1);
		}
		//get projection data
		if (Org.GetProjData(Buffer[0]) != 0)
		{
			fprintf(output,
					"\n **** ERROR: could not get projection data from file"
							"Org\n  **** program aborted\n\n");
		}

		// array set selection
		if (NewM.SelectProj(j) != 0)
		{
			fprintf(output, "\n **** ERROR: could not select projection of file"
					"NewM\n  **** program aborted\n\n");
			exit(-1);
		}

		//get projection data
		if (NewM.GetProjData(Buffer[1]) != 0)
		{
			fprintf(output,
					"\n **** ERROR: could not get projection data from file"
							"NewM\n  **** program aborted\n\n");
		}

		// array set selection
		if (NewP.SelectProj(j) != 0)
		{
			fprintf(output, "\n **** ERROR: could not select projection of file"
					"NewP\n  **** program aborted\n\n");
			exit(-1);
		}

		//get projection data
		if (NewP.GetProjData(Buffer[2]) != 0)
		{
			fprintf(output,
					"\n **** ERROR: could not get projection data from file"
							"NewP\n  **** program aborted\n\n");
		}

		//calculate corrected value for each ray in the projection

		for (k = 0; k < NoOfRays; k++)
		{

			REAL den = Buffer[2][k];
			REAL denden = 1.0;
			REAL newden = 0.0;
			for (i = 0; i <= noDegree; i++)
			{
				newden += denden * a[i];
				denden = denden * den;
			}

			Buffer[3][k] = miu * Buffer[0][k] + Buffer[1][k] - miu * newden;
		}

		//add corrected projection to the Corrected file
		if (Corrected.AppendProj(&ProjectionHeader[0], Buffer[3]) != 0)
		{
			fprintf(output, "\n **** ERROR: could not append projection to file"
					"Corrected\n  **** program aborted\n\n");
		}
	}

	//close the files
	Org.Close();
	NewM.Close();
	NewP.Close();
	Corrected.Close();

	//delete allocated memory
	if (Buffer[0] != NULL)
		delete[] Buffer[0];
	if (Buffer[1] != NULL)
		delete[] Buffer[1];
	if (Buffer[2] != NULL)
		delete[] Buffer[2];
	if (Buffer[3] != NULL)
		delete[] Buffer[3];
	delete Buffer;

	return 0;
}

