/*
 **********************************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/conv.cpp $
 $LastChangedRevision: 85 $
 $Date: 2014-07-02 16:07:08 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 **********************************************************************

 THIS IS A GENERAL IMPLEMENTATION OF THE THREE BEST KNOWN
 CONVOLUTION METHODS, NAMELY THOSE OF BRACEWELL + RIDDLE[1],
 RAMACHANDRAN + LAKSHMINARAYANAN[2], AND SHEPP + LOGAN[3].
 THE CONTROL CARD FOLLOWING THE EXECUTE CARD CONTAINS (IN THE
 SNARK FREE FORMAT) THE FOLLOWING INFORMATION:
 1) THE DESIRED FILTER ("BAND", "SINC", "COSINE" OR "HAMMING")
 2) THE CUTOFF PARAMETER (OR THE VALUE OF ALPHA FOR HAMMING)
 3) THE INTERPOLATION METHOD DESIRED
 THE BAND LIMITING FILTER CORRESPONDS TO THE FILTER USED BY
 BRACEWELL + RIDDLE(1) AND RAMACHANDRAN + LAKSHMINARAYANAN(2).
 THE SINC FILTER IS GENERALIZATION OF THE FILTER USED BY SHEPP
 + LOGAN(2).
 THE COSINE FILTER IS A COMMON LOWPASS FILTER APPLIED TO THE
 RECONSTRUCTION PROBLEM.
 THE HAMMING FILTER IS THE GENERALIZED HAMMMING FILTER, A SPECI
 CASE OF WHICH WAS USED BY CHESLER + REIDERER(4).
 THE CUTOFF VALUE REPRESENTS THE FRACTION OF FOURIER SPACE
 POWER TO INCLUDE IN THE RECONSTRUCTION.  WHEN 'CUTOFF' IS ZERO,
 IT IS TAKEN TO 1.0.  WHEN 'CUTOFF' IS NEGATIVE, IT IS SET
 TO '2.0*PRJNUM*PINC/(PI*NELEM*PIXSIZ)'.  IF 'CUTOFF' IS GREATER
 THAN 1.0, IT IS SET TO 1.0.
 IF 'UNI' = .TRUE. (UNIFORM SPACING SPECIFIED),
 THEN THE CUTOFF FORMS A CIRCULAR REGION IN FOURIER SPACE
 WRERE THE DIAMETER OF THE CIRCLE IS 'CUTOFF/PIXSIZ'.
 IF 'VRI' = .TRUE. THEN THE CUTOFF FORMS A SQUARE REGION WHERE
 THE LENGTH OF A SIDE IS 'CUTOFF/PIXSIZ'.
 EXAMPLE CONTROL CARD SEQUENCE
 EXECUTE CONVOLUTION
 BAND LIMITING CUTOFF = -1.0 INTERPOLATION = 2
 WARNING.....
 THIS VERSION OF THE CONVOLUTION METHOD WEIGHTS PROJECTION "NP"
 BY "(THETA(NP+1)-THETA(NP-1))/2.0".  IN ORDER TO ACCOMPLISH THIS
 IT IS ASSUMED THAT THE PROJECTIONS ARE ARRANGED IN INCREASING
 ORDER ON THETA AND THAT THETA(PRJNUM)-THETA(1) IS LESS THAN
 PI IN THE PARALLEL CASE OR TWOPI IN THE DIVERGENT CASE.

 REFERENCES:
 [1]  BRACEWELL + RIDDLE, INVERSION OF FAN-BEAM SCANS IN RADIO
 ASTROMOMY, ASTROPHYS. J., V.150, PP.427-434 (1967).
 [2]  RAMACHANDRAN + LAKSHMINARAYANAN, THREE-DIMENSIONAL
 RECONSTRUCTION FROM RADIOGRAPHS AND ELECTRON MICROGRAPHS:
 APPLICATION OF CONVOLUTIONS INSTEAD OF FOURIER TRANSFORMS,
 PROC. NAT. ACAD. SCI. U.S., V.68, PP.2236-2240 (1971).
 [3]  SHEPP + LOGAN, THE FOURIER RECONSTRUCTION OF A HEAD SECTION,
 IEEE TRANS. NUCL. SCI., NS-21, PP.21-43 (1974).
 [4]  CHESLER + RIEDERER, RIPPLE SUPPRESSION DURING RECONSTRUCTION
 IN TRANSVERSE TOMOGRAPHY, PHYS. MED. BIOL., V.20, PP.632-636
 (1975).
 */

#include <cstdlib>
#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "geom.h"
#include "spctrm.h"
#include "fourie.h"
#include "consts.h"
#include "uiod.h"
#include "int2str.h"
#include "anglst.h"
#include "bckprj.h"
#include "qinit.h"
#include "projfile.h"
#include "infile.h"

#include "conv.h"

BOOLEAN conv_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{

	INTEGER filter;

	// modified calls to getwrd() to use the 4-parameter version. Lajos, Dec 13, 2004
	static const INTEGER fname[4] =
	{
			CHAR2INT('b', 'a', 'n', 'd'),
			CHAR2INT('s', 'i', 'n', 'c'),
			CHAR2INT('c', 'o', 's', 'i'),
			CHAR2INT('h', 'a', 'm', 'm')
	};

	REAL thmod;
	REAL thfac;
	INTEGER word;
	BOOLEAN eol;
	INTEGER ndelta;
	INTEGER gpnum;
	REAL* g = NULL;        //wei 3.2005
	INTEGER qnum;
	REAL* q = NULL;        //wei 3.2005
	REAL* gp = NULL;        //wei 3.2005
	REAL theta;
	REAL sinth;
	REAL costh;
	REAL theta0;
	INTEGER np;
	REAL theta1;
	INTEGER np2;
	REAL theta2;
	REAL sinth2;
	REAL costh2;
	REAL w;
	INTEGER izro;
	REAL* gpi;
	INTEGER nr;
	REAL* gi;
	REAL* qi;
	REAL sum;
	INTEGER iz;
	INTEGER i;

	// INITIALIZATION

	if (iter > 1)
		return FALSE;
	thmod = Consts.pi;
	thfac = 2.0;
	if (!GeoPar.par)
	{
		thmod = Consts.twopi;
		thfac = 4.0;
	}

	word = InFile.getwrd(TRUE, &eol, fname, 4);
	for (filter = 0; filter < 4; filter++)
	{
		if (word == fname[filter])
			goto L30;
	}

	fprintf(output, "\n***** %s is not a valid filter name", int2str(word)); //(JD 12/31/03)

	return TRUE;

	L30: Fourie.cutoff = InFile.getnum(FALSE, &eol);
	Fourie.interp = InFile.getint(FALSE, &eol);

	if (eol)
	{
		fprintf(output, "\n *** not enough arguments");

		return TRUE;
	}

	if (Fourie.cutoff < 0.0)
		Fourie.cutoff = (REAL) 2.0 * GeoPar.prjnum * GeoPar.pinc
				/ ((REAL) Consts.pi * GeoPar.nelem * GeoPar.pixsiz);
	if (Fourie.cutoff <= Consts.zero)
		Fourie.cutoff = 1.0; // bug 151 - swr - 9/24/05
	if (Fourie.cutoff > 1.0)
		Fourie.cutoff = 1.0;

	fprintf(output, "\n          convolution reconstruction algorithm");
	fprintf(output, "\n          using filter %s", int2str(word));
	fprintf(output, "\n          cutoff (or alpha) = %10.4f", Fourie.cutoff);

	if ((Fourie.interp < -1) || (Fourie.interp > 6))
	{
		fprintf(output, "\n *** invalid interpolation %4i", Fourie.interp);

		return TRUE;
	}

	if (Fourie.interp > 0)
	{
		fprintf(output, "\n         %2i point lagrange interpolation",
				Fourie.interp);
	}
	if (Fourie.interp < 0)
	{
		fprintf(output, "\n          cubic spline interpolation");
	}
	if (Fourie.interp == 0)
	{
		fprintf(output, "\n          band limiting (sinc) interpolation");
	}

	ndelta = 2 * Fourie.interp;
	if (Fourie.interp < 0)
		ndelta = 6;
	gpnum = GeoPar.snrays + ndelta;

	// G IS THE BASE OF PROJECTION DATA

	g = new REAL[GeoPar.usrays];

	// Q IS THE BASE OF THE CONVOLUTING FUNCTION VALUES

	qnum = GeoPar.nrays + ndelta;

	q = new REAL[qnum];

	qinit(q, qnum, filter, Fourie.cutoff);

	// GP IS THE BASE OF THE CONVOLUTED PROJECTIONS

	gp = new REAL[gpnum];

	// BEGIN THE RECONSTRUCTION

	Anglst.getang(GeoPar.prjnum - 1, &theta, &sinth, &costh);

	theta0 = theta - thmod;
	for (np = 0; np < GeoPar.prjnum; np++)
	{

		ProjFile.ReadProj(np, g, GeoPar.usrays);

		Anglst.getang(np, &theta1, &sinth, &costh);

		// COMPUTE THE WEIGHT FOR THE PROJECTION

		np2 = np + 1;

		if (np == GeoPar.prjnum - 1)
			np2 = 0;

		Anglst.getang(np2, &theta2, &sinth2, &costh2);

		if (np == (GeoPar.prjnum - 1))
			theta2 += thmod;
		w = (theta2 - theta0) / thfac;
		if (w <= 0.0)
		{
			fprintf(output, "\n *** angles not in increasing order");

			if (g != NULL)
				delete[] g;
			if (q != NULL)
				delete[] q;
			if (gp != NULL)
				delete[] gp;         //wei 3/2005

			return TRUE;
		}

		Fourie.rinc = GeoPar.pinc;

		if (GeoPar.vri)
			Fourie.rinc = GeoPar.pinc * (REAL) MAX0(fabs(sinth), fabs(costh));
		if (GeoPar.div)
			Fourie.rinc = GeoPar.pinc * GeoPar.radius / GeoPar.stod;

		w /= Fourie.rinc;

		if (GeoPar.strip)
			w /= Fourie.rinc;

		theta0 = theta1;

		// COMPUTE CONVOLUTED PROJECTIONS

		izro = (GeoPar.usrays - gpnum) / 2;

		gpi = gp;

		for (nr = 0; nr < gpnum; nr++)
		{
			gi = g;
			qi = q + abs(izro);
			sum = 0.0;

			iz = MIN0(izro, GeoPar.usrays);
			if (iz > 0)
			{
				for (i = 0; i < iz; i++)
				{
					sum += *gi * *qi;
					gi++;
					qi--;
				}
			}

			izro++;
			iz = MAX0(izro, 1);
			if (iz <= GeoPar.usrays)
			{
				for (i = iz; i <= GeoPar.usrays; i++)
				{
					sum += *gi * *qi;
					gi++;
					qi++;
				}
			}

			*gpi++ = w * sum;
		}

		// BACK PROJECT THE DATA
		bckprj(recon, GeoPar.nelem, gp, gpnum, sinth, costh,
				GeoPar.pixsiz / Fourie.rinc, Fourie.interp);

	}

	if (g != NULL)
		delete[] g;
	if (q != NULL)
		delete[] q;
	if (gp != NULL)
		delete[] gp;         //wei 3/2005

	return FALSE;
}

