/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/sirt.cpp $
 $LastChangedRevision: 97 $
 $Date: 2014-07-02 20:07:41 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 IMPLEMENTATION OF THE SIRT ALGORITHM OF PETER GILBERT
 TOGETHER WITH EXTENSIONS AND MODIFICATIONS BY
 LAKSHMINARAYANAN AND LENT OF SUNY BUFFALO
 MAY, 1977;  REVISED JANUARY, 1978.
 SEE TECHNICAL REPORT 95 OF DEPT OF COMPUTER SCIENCE
 PROGRAMMED IN ANSI FORTRAN FOR PORTABILITY
 USES AUXILIARY FILES USER1 AND USER2 TO REDUCE CORE

 METHOD...CONTROLS ALGORITHM SELECTION
 IF 'GSIRT' USE LAK-LENT GENERALIZATION OF GILBERT SIRT
 SEE EQUATION (20) OF REPORT 95
 RHO(Q+1) = RHO(Q) + 1.0/SIGMA *(B-(DN)**-1 * PTRANS * P *
 RHO(Q)
 WHERE
 B = (DL)**-1 * P TRANS * R
 DN = P TRANS * P * U, U A VECTOR OF ONES
 DL IS A WEIGHTED SUM OF RAY LENGTHS
 R IS VECTOR OF PROJECTION DATA

 FOR 'LSIRT', MODIFIED SIRT ALGORITHMS OF LAK AND LENT
 ARE SELECTED
 GENERAL FORM OF MODIFIED SIRT IS (SEE EQUATION (23) )
 RHO(Q+1) = RHO(Q) + 1.0/SIGMA * (B - DV**-1 * PTRANS *
 Z * V**2 * P * RHO(Q))
 DEFINITIONS FOLLOW
 N(J) = NUMBER OF PIXELS IN RAY J
 L(J) = LENGTH OF RAY J
 W(J) = WIDTH OF RAY J
 FOR THE LSIRT ALGORITHMS, (Z * V**2)(J) TAKES THE FOLLOWING FORM
 N(J)**POWER2
 DV(I) TAKES THE FOLLOWING FORM
 SUM OVER ALL RAYS J PASSING THROUGH PIXEL I OF N(J)**POWER1
 *** DV IS DESIGNATED THROUGHOUT THIS PROGRAM BY THE SYMBOL DN ***
 B (IN THE TECHNICAL REPORT) HAS THE FOLLOWING FORM
 DV**-1 * PTRANS * Z * V * R
 FOR EACH RAY J
 (Z * V * R)(J) = R(J)/(L(J)*W(J)) * N(J)**POWER1
 VALUES OF LSIRT MODIFIER
 LSIRT = 1       POWER1=1            POWER2=0
 LSIRT = 2       POWER1 = 0          POWER2 = -1
 LSIRT = 3       POWER1 = -1         POWER2 = -2
 TO GET CONSTRAINED SIRT, USE THE SNARK MODE CARD

 ***INPUT SEQUENCE

 THE FOLLOWING SEQUENCE OF KEYWORDS IS AVAILABLE TO THE USER

 METHOD
 THIS KEYWORD MUST BE PRESENT
 IT MAY BE FOLLOWED BY EITHER 'GSIRT' OR 'LSIRT TYPE'.
 TYPE IS AN INTEGER MODIFIER WHICH MUST TAKE ON VALUES
 1, 2, OR 3.

 RELAX
 IF THIS IS PRESENT, IT MUST BE FOLLOWED BY THE FLOATING-
 POINT VALUE OF THE RELAXATION PARAMETER;
 SIGMA
 IF THIS IS PRESENT, IT MUST BE FOLLOWED BY THE FLOATING-
 POINT VALUE OF SIGMA;
 RELAXATION FACTOR = 1.0/SIGMA

 *****  RELAX AND SIGMA MUST NOT BOTH BE USED  *****

 IF NEITHER RELAX OR SIGMA IS USED, RELAXATION FACTOR =
 SIGMA = 1.0

 START
 MAY BE FOLLOWED BY 'BACKPROJECTED'
 THIS CAUSES A BACKPROJECTED IMAGE TO BE USED AS THE
 INITIAL GUESS
 IF START IS NOT PRESENT, THE EXEC COMMAND OPTION IS USED

 NORMAL
 THIS CAUSES ALL RECONSTRUCTIONS (OTHER THAN THE INITIAL GUESS)
 TO BE NORMALIZED TO AVEDEN.  THIS STEP IS OMITTED IF
 AVEDEN IS TOO SMALL OR IF THE TOTAL DENSITY OF THE
 ESTIMATED IMAGE IS LOW

 EXAMPLE INPUT SEQUENCES FOLLOW

 METHOD GSIRT SIGMA = .5 START NORMAL
 METHOD LSIRT 1 RELAX = 3.

 **** THIS SIRT IS FOR STRIP PROJECTION DATA ONLY  ****

 ADDED FORMAT 1090 FOR WRITING INTO USER1 BAND USER2..7/21/88.SADA
 */

#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "geom.h"
#include "spctrm.h"
#include "fourie.h"
#include "raysel.h"
#include "modefl.h"
#include "consts.h"
#include "uiod.h"
#include "raylen.h"
#include "posit.h"
#include "anglst.h"
#include "ray.h"
#include "projfile.h"
#include "infile.h"
#include "blob.h"

#include "sirt.h"

#define  nrdis(rinc) MIN0(((INTEGER)(size * (fabs(costh) + fabs(sinth))/(rinc) - 0.001)), limit)

INTEGER sirt_class::Init()
{

	hmeth = CHAR2INT('m', 'e', 't', 'h');
	hgsir = CHAR2INT('g', 's', 'i', 'r');
	hlsir = CHAR2INT('l', 's', 'i', 'r');
	hstar = CHAR2INT('s', 't', 'a', 'r');
	hnorm = CHAR2INT('n', 'o', 'r', 'm');
	hrela = CHAR2INT('r', 'e', 'l', 'a');
	hsigm = CHAR2INT('s', 'i', 'g', 'm');
	hback = CHAR2INT('b', 'a', 'c', 'k');

	w1 = NULL;
	w2 = NULL;
	w3 = NULL;
	gen_inv_dn = NULL;
	vector_b = NULL;
	pbase = NULL;
	sbase = NULL;             //wei, 3/2005

	return 0;
}

INTEGER sirt_class::Reset()
{
	if (w1 != NULL)
		delete[] w1;
	if (w2 != NULL)
		delete[] w2;
	if (w3 != NULL)
		delete[] w3;
	if (gen_inv_dn != NULL)
		delete[] gen_inv_dn;
	if (vector_b != NULL)
		delete[] vector_b;
	if (pbase != NULL)
		delete[] pbase;
	if (sbase != NULL)
		delete[] sbase;               //wei, 3/2005

	return 0;
}

BOOLEAN sirt_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{

	static REAL totden;
	REAL size;
	INTEGER i;
	INTEGER np;
	REAL theta;
	REAL sinth;
	REAL costh;
	INTEGER nr;
	INTEGER numb;
	REAL snorm;
	REAL ax;
	REAL ay;
	REAL cth;
	REAL sth;
	REAL wl;
	REAL raynar;
	REAL raylar;
	REAL dl;
	INTEGER nb;
	INTEGER k;
	REAL raysum;
	REAL w;
	REAL psum;
	REAL ronew;
	REAL ratio;

	static BOOLEAN normal;
	BOOLEAN bkstrt, gsirt, sirt, eol;

	INTEGER limit, first, last;
	INTEGER word;
	static INTEGER lsirt; // set static hstau
	REAL relax;
	static REAL sigma;

	INTEGER area;

	if (Blob.pix_basis)
	{
		area = GeoPar.area;
	}
	else
	{
		area = Blob.area;
	}

	// STATEMENT FUNCTION ENABLING TO DISCARD RAYS THAT DONT
	// INTERSECT THE RECONSTRUCTION REGION IN THE SUMMATIONS
	// DESCRIBED IN THE PREFACE

	if (iter <= 1)
	{

		//major changes in the input reading sequence, using new getwrd funcion. hstau 1/2004
		lsirt = 0;
		//for stripe mode only!
		if (GeoPar.line)
		{
			fprintf(output, "\n *** this sirt algorithm is for strip data only, try quad algorithm for line data**");
			return TRUE;
		}

		// SELECT THE SNARK OR USER SET OF RAYS
		limit = GeoPar.lusray - GeoPar.midray;

		if (GeoPar.snrays != GeoPar.usrays)
		{

			//  CHECK FOR SNARK OR USER
			if (!RaySel.useray)
			{

				// SNARK OPTION IN EFFECT

				fprintf(output, "\n          snark rays selected");

				limit = GeoPar.lsnray - GeoPar.midray;
			}
			else
			{
				// USER OPTION IN EFFECT
				fprintf(output, "\n          user rays selected");
			}
		}

		//input phase
		INTEGER exp0[1] =
		{ hmeth };
		word = InFile.getwrd(TRUE, &eol, exp0, 1);

		if (eol)
		{
			fprintf(output, "\n          **** error - keyword method is missing***");
			return TRUE;
		}

		INTEGER exp1[2] =
		{ hgsir, hlsir };
		word = InFile.getwrd(FALSE, &eol, exp1, 2);

		if (eol)
		{
			fprintf(output, "\n          **** error - one and only one of the following methods must be specified: gsirt or lsirt***");
			return TRUE;
		}

		gsirt = word == hgsir;
		sirt = word == hlsir;

		if (word == hgsir)
		{
			fprintf(output, "\n          gsirt method");
		}
		else
		{  //hlsir
			lsirt = InFile.getint(FALSE, &eol);
			if (eol)
			{
				fprintf(output, "\n          **** error - must specify lsirt type ***");
				return TRUE;
			}
			if (!(lsirt > 0 && lsirt < 4))
			{
				return TRUE;
			}
			fprintf(output, "\n          lsirt method");
		}

		//default values
		sigma = 1.0;
		relax = 1.0;
		bkstrt = FALSE;
		normal = FALSE;

		// optional inputs
		INTEGER exp2[4] =
		{ hrela, hsigm, hstar, hnorm };
		word = InFile.getwrd(FALSE, &eol, exp2, 4);

		if (eol)
			goto L30;
		// finished reading input

		if (word == hrela)
		{
			relax = InFile.getnum(FALSE, &eol);
			if (eol)
			{
				fprintf(output, "\n          **** error - must specify relax ***");
				return TRUE;
			}
			if (relax < Consts.zero)
			{
				fprintf(output, "\n          **** error - relax too small or negative ***");
				return TRUE;
			}
			sigma = (REAL) 1.0 / relax;

			INTEGER exp2a[2] =
			{ hstar, hnorm };
			word = InFile.getwrd(FALSE, &eol, exp2a, 2);
			if (word == hstar)
				goto L10;
			else if (word == hnorm)
				goto L20;
			else
				goto L30;
			// eol: finished
		}
		else if (word == hsigm)
		{
			sigma = InFile.getnum(FALSE, &eol);
			if (eol)
			{
				fprintf(output, "\n          **** error - must specify sigma ***");
				return TRUE;
			}
			if (sigma < Consts.zero)
			{
				fprintf(output, "\n          **** error - sigma too small or negative ***");
				return TRUE;
			}
			INTEGER exp2b[2] =
			{ hstar, hnorm };
			word = InFile.getwrd(FALSE, &eol, exp2b, 2);
			if (word == hstar)
				goto L10;
			else if (word == hnorm)
				goto L20;
			else
				goto L30;
			// eol: finished
		}

		else if (word == hstar)
		{
			L10: bkstrt = TRUE;
			INTEGER exp3[2] =
			{ hback, hnorm };
			word = InFile.getwrd(FALSE, &eol, exp3, 2);
			if (eol || word == hback)
				goto L30;
			else
				goto L20;
			//normal
		}
		else
		{
			L20: normal = TRUE;
		}

		L30:
		// ECHOING INPUT

		fprintf(output, "\n          sigma = inverse of relaxation = %10.4f\n",
				sigma);
		// removed from the echoing hstau.2003

		// NONCONVERGENCE WARNING ON SIGMA
		if (sigma < .4999)
		{
			fprintf(output, "\n ***small sigma value.  probable divergence.***"); //hstau
		}

		if (Modefl.lofl || Modefl.upfl)
		{
			fprintf(output, "\n          this run is constrained sirt");
		}
		else
		{
			fprintf(output, "\n          this run is unconstrained sirt");
		}

		if (normal)
		{
			fprintf(output, "\n          this run is normalized sirt");
		}
		else
		{
			fprintf(output, "\n          this run is unnormalized sirt");
		}

		// CHECK ON SMALL TOTAL DENSITY AND ISSUE WARNING
		totden = ((REAL) (GeoPar.area)) * GeoPar.aveden;

		if (fabs(totden) <= Consts.zero)
		{
			normal = FALSE;
			fprintf(output, "\n ***average density too small; no normalization***"); //hstau
		}

		if (bkstrt)
		{
			fprintf(output, "\n          starting picture is back-projected density \n");
		}
		else
		{
			fprintf(output, "\n          starting picture is as in execute command");
		}

		// END INPUT CHECK AND ECHO
		//
		// *       START WORK
		//

		w1 = new REAL[area];
		w2 = new REAL[area];
		w3 = new REAL[area];
		gen_inv_dn = new REAL[area];
		vector_b = new REAL[area];

		pbase = new REAL[GeoPar.nrays];

		sbase = new INTEGER[GeoPar.prjnum];

		size = GeoPar.nelem * GeoPar.pixsiz / (REAL) 2.0;
		Fourie.rinc = GeoPar.pinc;

		// GET DN AND B ON USER1
		// GENERATE DN AND DL
		// IF 'LSIRT' DONT GENERATE DL,
		// GET P TRANSPOSE N*R/L DIRECTLY
		// WRITTEN FOR VARIABLE RAY WIDTH

		for (i = 0; i < area; i++)
		{
			w2[i] = 0.0;
			w1[i] = 0.0;
		}

		for (np = 0; np < GeoPar.prjnum; np++)
		{
			Anglst.getang(np, &theta, &sinth, &costh);
			if (sirt)
			{
				ProjFile.ReadProj(np, pbase, GeoPar.nrays);
			}

			if (GeoPar.vri)
				Fourie.rinc = GeoPar.pinc
						* (REAL) MAX0(fabs(costh), fabs(sinth));

			first = GeoPar.midray - nrdis(Fourie.rinc);

			sbase[np] = first;
			//last = GeoPar.nrays + 1 - first;
			last = GeoPar.nrays - 1 - first; //should be 2 less. hstau

			for (nr = first; nr <= last; nr++)
			{
				if (Blob.pix_basis)
				{
					ray(np, nr, list, weight, &numb, &snorm);
				}
				else
				{
					Blob.bray(np, nr, list, weight, &numb, &snorm);
				}
				if (numb != 0)
				{
					posit(np, nr, &ax, &ay, &cth, &sth);
					wl = Fourie.rinc
							* raylen(SOT_rect, ax, ay, cth, sth, 0., 0., size,
									size, 1., 0.);
					if (!sirt)
					{
						raynar = (REAL) (numb);
						raylar = wl;
					}
					else
					{
						dl = pbase[nr] / wl;
						switch (lsirt)
						{
						case 1:
							raynar = float(numb);
							raylar = dl * raynar;
							break;

						case 2:
							raynar = 1.0;
							raylar = dl;
							break;

						case 3:
							raynar = ((REAL) 1.0) / numb;
							raylar = dl * raynar;
							break;
						}
					}

					for (nb = 0; nb < numb; nb++)
					{
						k = list[nb];
						w2[k] += raynar;
						w1[k] += raylar;
					}
				}
			}

		}

		if (!sirt)
		{

			// GSIRT
			// TRYING FOR COMPATIBILITY IN CASE SOME PIXELS DO NOT
			// CONTRIBUTE TO THE PROJECTION DATA

			for (i = 0; i < area; i++)
			{
				if (w1[i] == 0.0)
					w2[i] = 0.0;
				if (w2[i] > Consts.zero)
					w2[i] = ((REAL) 1.0) / w2[i];
				if (w2[i] == 0.0)
					w1[i] = 0.0;
			}

			// W2 NOW HOLDS GENERALIZED INVERSE OF DN
			// WORK HOLDS DL

			for (i = 0; i < area; i++)
			{
				gen_inv_dn[i] = w2[i];
			}

			// GET P TRANSPOSE R

			for (i = 0; i < area; i++)
			{
				w2[i] = 0.0;
			}

			for (np = 0; np < GeoPar.prjnum; np++)
			{
				ProjFile.ReadProj(np, pbase, GeoPar.nrays);

				first = sbase[np];
				last = GeoPar.nrays - 1 - first;  //should be 2 less. hstau


				for (nr = first; nr <= last; nr++)
				{
					raysum = pbase[nr];
					if (Blob.pix_basis)
					{
						ray(np, nr, list, weight, &numb, &snorm);
					}
					else
					{
						Blob.bray(np, nr, list, weight, &numb, &snorm);
					}
					if (numb != 0)
					{
						for (nb = 0; nb < numb; nb++)
						{
							k = list[nb];
							w2[k] += raysum;
						}
					}
				}

			}

			for (i = 0; i < area; i++)
			{
				w = w1[i];
				if (w == 0.0)
					w2[i] = 0.0;
				if (w > Consts.zero)
					w2[i] /= w;
			}
		}
		else
		{

			// LSIRT OPTIONS

			// WORK HOLDS PART OF B; GET GENERALIZED INVERSE IN w2

			for (i = 0; i < area; i++)
			{
				if (w2[i] > Consts.zero)
					w2[i] = ((REAL) 1.0) / w2[i];
			}

			for (i = 0; i < area; i++)
			{
				gen_inv_dn[i] = w2[i];
			}

			// SET UP B VECTOR FOR 'LSIRT'

			for (i = 0; i < area; i++)
			{
				w2[i] = w1[i] * w2[i];
			}
		}

		for (i = 0; i < area; i++)
		{
			vector_b[i] = w2[i];
		}

		if (bkstrt)
		{
			for (i = 0; i < area; i++)
				recon[i] = w2[i];
		}

		// ITERATIONS BEGIN HERE
	}

	for (i = 0; i < area; i++)
	{
		w1[i] = vector_b[i];
	}

	// WORK HOLDS B

	for (i = 0; i < area; i++)
	{
		w1[i] = recon[i] + w1[i] / sigma;
	}

	for (i = 0; i < area; i++)
	{
		w3[i] = w1[i];
	}

	// GOING TO GENERATE P TRANSPOSE P RHO IN WORK

	for (i = 0; i < area; i++)
	{
		w1[i] = 0.0;
	}

	for (np = 0; np < GeoPar.prjnum; np++)
	{
		first = sbase[np];
		last = GeoPar.nrays - 1 - first;    // should be 2 less. hstau


		for (nr = first; nr <= last; nr++)
		{

			if (Blob.pix_basis)
			{
				ray(np, nr, list, weight, &numb, &snorm);
			}
			else
			{
				Blob.bray(np, nr, list, weight, &numb, &snorm);
			}

			if (numb != 0)
			{
				psum = 0.;
				for (nb = 0; nb < numb; nb++)
				{
					k = list[nb];
					psum += recon[k];
				}

				if (lsirt == 2)
					psum /= float(numb);
				if (lsirt == 3)
					psum /= float(numb * numb);

				for (nb = 0; nb < numb; nb++)
				{
					w1[list[nb]] += psum;
				}
			}
		}
	}

	for (i = 0; i < area; i++)
	{
		recon[i] = gen_inv_dn[i];
	}


	// RECON HOLDS GENERALIZED INVERSE OF DN
	// WORK HOLDS PTRANSPOSE P RHO

	for (i = 0; i < area; i++)
	{
		w1[i] *= recon[i];
	}

	for (i = 0; i < area; i++)
	{
		recon[i] = w3[i];
	}

	psum = 0.;
	for (i = 0; i < area; i++)
	{
		ronew = recon[i] - w1[i] / sigma;
		if (Modefl.lofl)
			ronew = MAX0(ronew, Modefl.lower);
		if (Modefl.upfl)
			ronew = MIN0(ronew, Modefl.upper);
		recon[i] = ronew;

		if (!Blob.pix_basis)
		{  //if blobs
			if (i % 2 == Blob.pr)
			{
				psum += ronew;
			}
		}
		else
		{
			psum += ronew;
		}
	}

	if (!normal)
	{

		return FALSE;
	}

	if (fabs(psum) > Consts.zero)
	{
		ratio = totden / psum;

		for (i = 0; i < area; i++)
		{
			recon[i] *= ratio;
		}

		return FALSE;
	}

	fprintf(output,
			"\n ***current total density too small; no normalization for this iteration***"); //hstau

	return FALSE;
}
