/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/trm5.cpp $
 $LastChangedRevision: 154 $
 $Date: 2014-07-28 11:32:36 -0400 (Mon, 28 Jul 2014) $
 $Author: olangthaler $
 ***********************************************************

 Kullback-Leibler value based stopping criterion

 */

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <math.h>
#include <limits>
#include "trm5.h"
#include "anglst.h"
#include "projfile.h"
#include "pseudo.h"
#include "consts.h"
#include "infile.h"
#include "blob.h"

using namespace std;

void trm5_class::Init()
{
	// read desired epsilon from command
	BOOLEAN eol;
	epsilon = InFile.getnum(FALSE, &eol);

	if (epsilon <= Consts_class::zero)
	{
		cout << "\n **** epsilon must be specified and greater than ZERO";
		cout << "\n **** program aborted\n";
		exit(-1);
	}

	// read reporting command (disabled by default)
	INTEGER rprt = CHAR2INT('r', 'p', 'r', 't');
	if (InFile.getwrd(FALSE, &eol, &rprt, 1) == rprt)
	{
		report = true;
		cout << "\n         reporting is enabled";
		cout << "\n         reporting file: RPRTklds";

		skips = InFile.getnum(FALSE, &eol);
		if (skips > 0)
		{
			cout << "\n         report skip factor " << skips;
		}
		else
		{
			cout << "\n         reporting on every iteration";
		}
	}
	else
	{
		report = false;
		cout << "\n         reporting is disabled";
	}

	cout << "\n         epsilon = " << epsilon;
}

BOOLEAN trm5_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{
	REAL* reconp = recon;
	// if currently blob, pixelize it
	if (!Blob.pix_basis)
	{
		reconp = new REAL[GeoPar.area];
		Blob.blob2pix(recon, reconp);
	}

	// variables needed during computation of the current epsilon
	INTEGER numb;
	REAL snorm;
	REAL raysum;
	REAL psum;
	BOOLEAN continu=true;

	// compute current KL distance
	REAL currentEpsilon = 0;
	for (INTEGER i = 0; i < GeoPar.prjnum && continu; i++)
	{
		for (INTEGER j = GeoPar.fusray; j <= GeoPar.lusray; j++)
		{
			raysum = Anglst.prdta(i, j);
			psum = pseudo(reconp, i, j, list, weight, &numb, &snorm, TRUE, TRUE);

			if (numb != 0)
			{
				if (raysum<0)
				{
					// report on non-applicability and abort
					std::cout << "\n****Error: Negative raysum in prjfil detected, KLDS is not applicable! Terminating algorithm execution.";
					return true;
				}
				else if (!isnan(currentEpsilon) && currentEpsilon!=std::numeric_limits<double>::infinity() && currentEpsilon!=std::numeric_limits<double>::max())
				{
					if (raysum<=Consts.zero && psum<=Consts.zero)
					{
						// do nothing
					}
					else if (raysum<=Consts.zero)
					{
						currentEpsilon += psum;
					}
					else if (psum<=Consts.zero)
					{
						// set to infinity and terminate
						if (std::numeric_limits<double>::has_infinity) currentEpsilon = std::numeric_limits<double>::infinity();
						else currentEpsilon = std::numeric_limits<double>::max();

						continu=false;
						break;
					}
					else
					{
						if ((raysum / psum) > Consts.zero) currentEpsilon += (raysum * log(raysum / psum));
						else currentEpsilon += (raysum * log(Consts.zero));

						currentEpsilon += psum - raysum;
					}
				}
			}
		}
	}
	if (!Blob.pix_basis)
	{
		delete[] reconp;
	}

	if (currentEpsilon <= epsilon)
	{
		// additional output
		if (report)
		{
			cout << "\n         current epsilon (KL distance) = " << currentEpsilon;

			FILE* RPRTfile;
			if (iter==1)
			{
				RPRTfile = fopen("RPRTklds", "w");
				fprintf(RPRTfile,"Kullback-Leibler distance stopping criterion reporting output");
				fprintf(RPRTfile,"\n    iter      Kullback-Leibler distance");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, currentEpsilon);
			}
			else
			{
				RPRTfile = fopen("RPRTklds", "a");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, currentEpsilon);
			}
			fclose(RPRTfile);
		}

		return true;
	}
	else
	{
		// additional output
		if (report && (iter == 1 || skips == 0 || (iter % (INTEGER) skips) == 0))
		{
			cout << "\n         current epsilon (KL distance) = " << currentEpsilon;

			FILE* RPRTfile;
			if (iter==1)
			{
				RPRTfile = fopen("RPRTklds", "w");
				fprintf(RPRTfile,"Kullback-Leibler distance stopping criterion reporting output");
				fprintf(RPRTfile,"\n    iter    Kullback-Leibler distance");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, currentEpsilon);
			}
			else
			{
				RPRTfile = fopen("RPRTklds", "a");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, currentEpsilon);
			}
			fclose(RPRTfile);
		}

		return false;
	}
}
