/** @file SnarkDisplay.cpp
    @package snark14Display
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
    
    modification: Joanna Klukowska 1/13/2009
	    added choice of grayScale display for plot
	    in addition to color display
// 
 */

// sd headers
#include "SnarkDisplay.hpp"
#include "displaylines.hpp"
#include "displayprojection.hpp"
#include "eval_helper.hpp"
#include "lines.hpp"
#include "line_real_t.hpp"
#include "plot_t.hpp"
#include "linesimages.hpp"
#include "proj.hpp"
#include "displaywindow.hpp"
#include "selectEvalExecutions.hpp"
#include "variables.hpp"
#include "range.hpp"
#include "openimages.hpp"

// C++ headers
#include <iostream>

// C-like headers
#include <cstdio>
#include <ctype.h>
#include <malloc.h>

// Qt headers
#include <qapplication.h>
#include <qdialog.h>
#include <qfiledialog.h>
#include <qframe.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qpen.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qmessagebox.h>
#include <qregexp.h> 

const char *myTag = "sd140707";

/* 
 *  Constructs a SnarkDisplay which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'fl'
 */
SnarkDisplay::SnarkDisplay(QWidget* parent, const char* name, WFlags fl)
: QWidget(parent, name, fl) {
    std::cout << "Executing snark14Display - " << myTag << std::endl;
    numimages = 0;
    desk = QApplication::desktop();
    maxSizex = desk->size().width();
    maxSizey = desk->size().height();
    if (!name)
        setName("snark14Display");
    resize(250, 400);
    setSizePolicy(QSizePolicy((QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, sizePolicy().hasHeightForWidth()));
    setMinimumSize(QSize(250, 400));
    setMaximumSize(QSize(250, 400));
    setCaption(tr("snark14Display"));

    recfilbutton = new QPushButton(this, "recfilbutton");
    recfilbutton->setGeometry(QRect(40, 25, 170, 40));
    recfilbutton->setText(tr("Open Recfil File"));
    QObject::connect(recfilbutton, SIGNAL(clicked()), this, SLOT(openRecfilFile()));

    imagesbutton = new QPushButton(this, "imagesbutton");
    imagesbutton->setGeometry(QRect(40, 75, 170, 40));
    imagesbutton->setText(tr("Open Images"));
    QObject::connect(imagesbutton, SIGNAL(clicked()), this, SLOT(callOpenImages()));

    rangebutton = new QPushButton(this, "rangebutton");
    rangebutton->setGeometry(QRect(40, 125, 170, 40));
    rangebutton->setText(tr("Set Global Features"));
    QObject::connect(rangebutton, SIGNAL(clicked()), this, SLOT(callRange()));

    linesbutton = new QPushButton(this, "linesbutton");
    linesbutton->setGeometry(QRect(40, 175, 170, 40));
    linesbutton->setText(tr("Display Lines"));
    QObject::connect(linesbutton, SIGNAL(clicked()), this, SLOT(callLines()));

    projfilbutton = new QPushButton(this, "projfilbutton");
    projfilbutton->setGeometry(QRect(40, 225, 170, 40));
    projfilbutton->setText(tr("Open Prjfil File"));
    QObject::connect(projfilbutton, SIGNAL(clicked()), this, SLOT(openPrjfilFile()));

    evalbutton = new QPushButton(this, "evalbutton");
    evalbutton->setGeometry(QRect(40, 275, 170, 40));
    evalbutton->setText(tr("Open Eval File"));
    QObject::connect(evalbutton, SIGNAL(clicked()), this, SLOT(openEvalFile()));

    Line1 = new QFrame(this, "Line1");
    Line1->setGeometry(QRect(0, 325, 260, 16));
    Line1->setFrameStyle(QFrame::HLine | QFrame::Sunken);

    quitbutton = new QPushButton(this, "quitbutton");
    quitbutton->setGeometry(QRect(10, 350, 100, 40));
    quitbutton->setText(tr("Quit"));
    QObject::connect(quitbutton, SIGNAL(clicked()), qApp, SLOT(quit()));

    helpbutton = new QPushButton(this, "helpbutton");
    helpbutton->setGeometry(QRect(140, 350, 100, 40));
    helpbutton->setText(tr("About"));
    QObject::connect(helpbutton, SIGNAL(clicked()), this, SLOT(about()));

    allocSnarkimages(numimages);
    //  allocDisplaywindows(numimages);
    displaywindowset = 0;
    displaywindowactive = 0;
    myRangeWinP = 0;
    myOpenImagesWinP = 0;
} // --SnarkDisplay ctor

/** Destroys the object and frees any allocated resources
 */
SnarkDisplay::~SnarkDisplay() {
    // no need to delete child widgets, Qt does it all for us
}

/** Calls file dialog and opens recfil file
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::openRecfilFile() {
    unsigned int asn, an;
    int i, j, ni, widthproj, widthexec;
    char name[256], oldname[256], nameexecution[13], namephantom[41];
    char namealg[5], nameprojection[13];
    const char *na;
    const DIGDimensions* Dimensions;
    QString itname, recfilfilename, s;
    QFileDialog fd(QString::null, "Recfil files (*recfil*)\n All files (*)", this, "Open Recfil File", true);
    fd.setCaption("Open Recfil File");
    fd.setSelection(QDir::currentDirPath());
    if (fd.exec() == QDialog::Accepted)
        recfilfilename = fd.selectedFile();
    else
        recfilfilename = "";
    if (openrecfil) {
        digrecfil.Close();
        free(settitles);
        for (i = 0; i < numimages; i++) {
            if (imageindexes[i]) {
                free(imageindexes[i]);
                imageindexes[i] = 0;
            }
            imagetitles[i] = 0;
        }
        if (myRangeWinP) {
            myRangeWinP->closeWindow();
            myRangeWinP = 0;
        }
        if (myOpenImagesWinP) {
            delete myOpenImagesWinP;
            myOpenImagesWinP = 0;
        }
        free(imageindexes);
        freeSnarkimages(numimages);
        freeDisplaywindows(numimages);
    }
    if (!recfilfilename.isEmpty()) {
        snprintf(name, sizeof(name)-1, "%s", recfilfilename.latin1());
        if (digrecfil.Open(name) != 0) {
            s.sprintf("Error!\nUnable to open file:\n%s\n", recfilfilename.latin1());
            QMessageBox::information(this, "SnarkDisplay", s);
            digrecfil.Close();
            return;
        } else {
          snprintf(oldname, sizeof(oldname)-1, "%s", name);
            phantomexists = false;
            if (digrecfil.GetDimensions(&Dimensions) != 0) {
                s.sprintf("%s is not a snark14 recfil\n", recfilfilename.latin1());
                QMessageBox::information(this, "SnarkDisplay", s);
                return;
            }
            if ((Dimensions->x != 1) && (Dimensions->y != 1) && (Dimensions->z != 1)) {
                s.sprintf("%s is not a snark14 recfil\n", recfilfilename.latin1());
                QMessageBox::information(this, "SnarkDisplay", s);
                return;
            }
            if (Dimensions->z == 1) {
                Sizex = Dimensions->x;
                Sizey = Dimensions->y;
            } else {
                if (Dimensions->y == 1) {
                    Sizex = Dimensions->x;
                    Sizey = Dimensions->z;
                } else {
                    Sizex = Dimensions->y;
                    Sizey = Dimensions->z;
                }
            }
            digrecfil.GetDataFormat(&dataformat);
            nameproj = (const char *) malloc(sizeof (const char) *81);
            digrecfil.GetTitle(&nameproj);
            strncpy(nameprojection, nameproj, 12);
            nameprojection[12] = 0;
            widthproj = 12;
            for (j = 11; j >= 0; j--) {
                if (nameprojection[j] == ' ' || nameprojection[j] == 0 || nameprojection[j] == 10) {
                    nameprojection[j] = 0;
                    widthproj--;
                } else
                    break;
            }
            digrecfil.GetNoOfArraySets(&asn);
            settitles = (const char **) malloc(sizeof (const char *) * asn);
            numimages = 0;
            for (i = 0; i < (static_cast<int> (asn)); i++) {
                digrecfil.SelectArraySet(i);
                digrecfil.GetArraySetTitle(&settitles[i]);
                digrecfil.GetNoOfArrays(&an);
                numimages += an;
            }
            imagetitles = (char **) malloc(sizeof (char *) * numimages);
            for (i = 0; i < numimages; i++) {
                imagetitles[i] = (char *) malloc(sizeof (char) *41);
            }
            imageindexes = (unsigned int **) malloc(sizeof (unsigned int *) * numimages);
            for (i = 0; i < numimages; i++) {
                imageindexes[i] = (unsigned int *) malloc(sizeof (unsigned int *) *2);
            }
            ni = 0;
            for (i = 0; i < (static_cast<int> (asn)); i++) {
                widthexec = 12;
                digrecfil.SelectArraySet(i);
                digrecfil.GetNoOfArrays(&an);
                digrecfil.GetArraySetType(&na);
                strncpy(namealg, na, 4);
                namealg[4] = 0;
                for (j = 0; j < 4; j++)
                    namealg[j] = toupper(namealg[j]);
                if (!strcmp(namealg, "PHAN")) {
                    phantomexists = true;
                    strncpy(namephantom, settitles[i], 40);
                    namephantom[40] = 0;
                    if (!strcmp(namephantom, ""))
                        itname.sprintf("_");
                    else
                        itname.sprintf("%-40s", namephantom);
                    snprintf(imagetitles[ni], sizeof(imagetitles[ni])-1, "%s", itname.latin1());
                    imageindexes[ni][0] = i;
                    imageindexes[ni][1] = 0;
                    ni++;
                } else {
                    strncpy(nameexecution, settitles[i], 12);
                    nameexecution[12] = 0;
                    widthexec = 12;
                    for (j = 11; j >= 0; j--) {
                        if (nameexecution[j] == ' ' || nameexecution[j] == 0) {
                            nameexecution[j] = 0;
                            widthexec--;
                        } else
                            break;
                    }
                    for (j = 0; j < (static_cast<int> (an)); j++) {
                        if (strcmp(nameprojection, ""))
                            itname.sprintf("%-*s_%-*s_%s_%04d_r_a", widthproj, nameprojection, widthexec, nameexecution, namealg, j + 1);
                        else
                            itname.sprintf("__%-*s_%4s_%04d_r_a", widthexec, nameexecution, namealg, j + 1);
                        snprintf(imagetitles[ni], sizeof(imagetitles[ni])-1, "%s", itname.latin1());
                        imageindexes[ni][0] = i;
                        imageindexes[ni][1] = j;
                        ni++;
                    }
                }
            } // --for on asn
            allocSnarkimages(numimages);
            allocDisplaywindows(numimages);
            if (openrecfil) {
                if (myOpenImagesWinP) {
                    myOpenImagesWinP->setImagesNames();
                }
            } else
                openrecfil = true;
        }
    }
} //--SnarkDisplay::openRecfilFile()

/** Starts a window for selecting images to be shown
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::callOpenImages() {
    QPoint topleft;
    topleft = this->pos();
    if (!openrecfil) {
        QMessageBox::information(this, "SnarkDisplay", "Error!\n"
                "No recfil file is open.");
    } else {
        //  noiw=new openimageswindow(desk,"Open Images",false,0,topleft.x()+30,topleft.y()+30);
        if (!myOpenImagesWinP) {
            myOpenImagesWinP = new openimageswindow(displaywindowset, desk, "Open Images", false, 0, topleft.x() + 30, topleft.y() + 30);
            //      myOpenImagesWinP = new openimageswindow(displaywindowset,this,"Open Images",false,0,topleft.x()+30,topleft.y()+30);
        }
        myOpenImagesWinP->hide(); // hack to give focus, setFocus() didn't work
        myOpenImagesWinP->show();
    }
} // --SnarkDisplay::callOpenImages()

/** Starts a lines window
@param void
@author Bruno M. Carvalho, deniz
@version 1.4
@comments majorly shortened due to new (Deniz'04) classes
 */
void SnarkDisplay::callLines() {
    if (std::verbose >= 2) {
        std::cout << "++SnarkDisplay::callLines()" << std::endl;
    }
    int l[4];
    int linetype, c, i, j, k, minx, maxx;
    int *siarray;
    int topleftx, toplefty;
    double **img;
    double miny, maxy;
    QString title, xaxis, yaxis, temp;
    QFont f;
    QPoint topleft;
    if (!openrecfil)
        QMessageBox::information(this, "SnarkDisplay", "Error!\n"
            "No recfil file is open.");
    else {
        if (!imageset[0]->isLoaded())
            imageset[0]->readImage(0, dataformat);
        linesimageswindow* nw = new linesimageswindow(this, "Multiple Images' Lines", true);
        c = nw->exec();
        if (c == QDialog::Accepted) {
            linetype = nw->getLinetype(); // row or column.  0 for row, 1 for column.
            l[0] = nw->getLine1();
            l[1] = nw->getLine2();
            l[2] = nw->getLine3();
            l[3] = nw->getLine4();
            title = ((nw->getTitle()).simplifyWhiteSpace());
            xaxis = nw->getXAxis();
            yaxis = nw->getYAxis();
            minx = nw->getMinX();
            maxx = nw->getMaxX();
            miny = nw->getMinY();
            maxy = nw->getMaxY();
            topleft = this->pos();
            topleftx = topleft.x();
            toplefty = topleft.y();
            siarray = nw->getSelectedImages();
            // (linetype==0) ==> show for suchandsuch row(s) all column values based on xrange
            // (linetype==1) ==> show for suchandsuch col(s) all row    values based on xrange
            QString rowOrColumn = ((linetype) ? ("Column ") : ("Row "));
            QString vs = ((linetype) ? ("Rows") : ("Columns"));
            for (k = 1; k <= siarray[0]; k++) {
                if (!imageset[siarray[k]]->isLoaded())
                    imageset[siarray[k]]->readImage(k, DIGDataFormat_ASCII);
            }
            // j is index for user-selected (keyed in) rows or columns
            for (j = 0; j < 4; j++) {
                // l[j] is the user-selected (keyed in) row or column "at position j"
                if (l[j] >= 0 && l[j] < Sizex) { // plot only if within range.
                    line_real_set_t myLRS(minx, maxx, miny, maxy); // create a brand new LRS for brand new plot
                    myLRS.clear(); // and be sure of it!
                    QString whichRowOrCol;
                    whichRowOrCol.setNum(l[j]);
                    QString plotName = title.simplifyWhiteSpace();
//                    plotName += " ::multiple images':: ";
                    plotName += " :: ";
                    plotName += (rowOrColumn + whichRowOrCol);
                    topleftx += 30;
                    toplefty += 30;
                    QPainter picasso;
                    QPixmap* pmp = new QPixmap(sd_line_t::WIDTH, sd_line_t::HEIGHT);
                    pmp->fill(Qt::white);
                    picasso.begin(pmp);
                    picasso.setBackgroundColor(Qt::white);
                    for (k = 1; k <= siarray[0]; k++) { // k is array index for the image within recfil
                        QString LRName = imagetitles[siarray[k]]; // something like "brainscan_ART_iter_5"
                        line_real_t myLR(LRName);
                        if (!(imageset[siarray[k]]->isLoaded())) {
                            imageset[siarray[k]]->readImage(k, DIGDataFormat_ASCII);
                        }
                        img = imageset[siarray[k]]->getImage();
                        int& row = ((linetype) ? (i) : (l[j])); // not assignment, but rather, unification.
                        int& col = ((linetype) ? (l[j]) : (i)); // (linetype==1) means l[j] is col, else l[j] is row
                        for (i = 0; i < Sizey; i++) { // recfils are Sizey by Sizey
                            myLR.add(i, img[row][col]); // x is independent i, y is dependent img[row][col]
                        }
                        myLRS.add(myLR); // add my LineReal due to selected row/col in this (k) image to my LRS
                    } // --for each image k (plot all of them on the same thing.)
                    plot_t myPlot(myLRS); // plot myLRS due to single selection of row/col (up to 4)
                    myPlot.setName(plotName);
                    myPlot.setXAxisName(xaxis);
                    myPlot.setYAxisName(yaxis);
                    // jk 1/13/2009 adding grayScale option for graphs
                    if (nw->isGrayScaleChecked()) { // picasso paints on QPixmap pointed to by pmp
                        myPlot.plotGray(picasso);
                    } else {
                        myPlot.plot(picasso);
                    }

                    if (toplefty > (maxSizey - 100))
                        toplefty = 30;
                    // spawn a display for each row/col plot, each plot potentially due to many images
                    displaylineswindow* nieuwDLWp = new displaylineswindow(desk, plotName, pmp, false, 0, topleftx, toplefty);
                    nieuwDLWp->show();
                } // --if of chosen row or column being within the appropriate range (i think)
            } // --for each chosen row or column (max. 4 of them)
        } // --"if" of Accepted QDialog
    } // --"else" of "no recfil is open"
} // --SnarkDisplay::callLines()

/** Starts a range window
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::callRange() {
    bool imageopened = false;
    int i;
    if (!openrecfil)
        QMessageBox::information(desk, "SnarkDisplay", "Error!\n"
            "No recfil file is open.");
    else {
        for (i = 0; i < numimages; i++) {
            if (displaywindowactive[i]) {
                imageopened = true;
            }
        }
        if (imageopened) {
            //      rangewindow* nw=new rangewindow(this,"Set Range",false,Qt::WType_TopLevel);
            if (myRangeWinP) {
                myRangeWinP->closeWindow();
                myRangeWinP = 0;
            }
            myRangeWinP = new rangewindow(this, "Set Range", false, Qt::WType_TopLevel);
            myRangeWinP->show();
        } else
            QMessageBox::information(desk, "SnarkDisplay", "Error!\n"
                "No image is open.");
    }
} // --SnarkDisplay::callRange()

/** Calls file dialog and opens PROJECTION file
    @param void
    @author Bruno M. Carvalho
    @version 1.0 */
void SnarkDisplay::openPrjfilFile() {
    QString projfilfilename, s;
    QPoint topleft;
    //  const DIGDimensions* Dimensions;
    char name[256];

    QFileDialog fd(QString::null, "Projfil files (*prjfil*)\n All files (*)", this, "Open Prjfil File", TRUE);
    fd.setCaption("Open Projfil File");
    fd.setSelection(QDir::currentDirPath());
    if (fd.exec() == QDialog::Accepted) {
        projfilfilename = fd.selectedFile();
        snprintf(name, sizeof(name)-1, "%s", projfilfilename.latin1());
        if (openprojfil) //*(JD 1/28/04)
            digprojfil.Close(); //(JD 1/28/04)
        if (digprojfil.Open(name) != 0) {
            s.sprintf("Error!\nUnable to open file:\n%s\nMaybe it is not a snark14 prjfil file!", projfilfilename.latin1());
            QMessageBox::information(this, "SnarkDisplay", s);
        } else {
            //      digprojfil.GetDataFormat(&projformat);
            projname = (const char *) malloc(sizeof (const char) *81);
            digprojfil.GetTitle(&projname);
            //      digprojfil.GetNoOfArraySets(&prjnum);
            digprojfil.GetNoOfProjs(&prjnum);
            //      digprojfil.GetDimensions(&Dimensions);
            //      usrays = Dimensions->x;
            digprojfil.GetDimensions(&usrays);
            projfil = new Snarkproj();
            // open prjfil in displaywindow
            topleft = this->pos();
            projwin = new displayprojection(desk, projname, false, 0, topleft.x(), topleft.y());
            projwin->show();
            projwin->resetscroll();
        }
        digprojfil.Open(name);
        openprojfil = true; //*(JD 1/28/04)
    }
}

/** Calls file dialog and opens eval file, throwing all information into global myEES
@param void
@author Bruno M. Carvalho for version 1.0, Deniz Sarioz after
@version 2.0 */
void SnarkDisplay::openEvalFile() {
    // myEES stands for "my eval_execution_set" defined in variables.h
    // eval_execution_set is defined in eval_helper.hpp and fleshed out in eval_helper.cpp
    myEES.clear(); // just in case
    // opening new eval file: get rid of ALL previous eval-related information
    QString evalfilename, s;
    FILE *fp;
    const unsigned maxlinelength = 512;
    const unsigned maxnamelength = 512;
    char line[maxlinelength]; // buffer for reading in lines
    char name[maxnamelength]; // buffer for reading in name
    memset(&(line[0]), 0, maxlinelength); // init line with null terminator
    memset(&(name[0]), 0, maxnamelength); // init name with null terminator
    int tmpint; // complete dummy
    QFileDialog fd(QString::null, "Eval files (*eval*)\n All files (*)", this, "Open Eval File", TRUE);
    fd.setCaption("Open Eval File");
    fd.setSelection(QDir::currentDirPath());
    if (fd.exec() == QDialog::Accepted) {
        evalfilename = fd.selectedFile();
        snprintf(name, sizeof(name)-1, "%s", evalfilename.latin1());
        if (std::verbose >= 2) printf("\nname of eval file is:\n%s\n", name);
        if ((fp = fopen(name, "r")) == NULL) {
            s.sprintf("Error!\nUnable to open file:\n%s\n", evalfilename.latin1());
            QMessageBox::information(this, "SnarkDisplay", s);
        } else {

            enum status_t {
                JUSTBEGAN, NEWEVALUATE, GLOBAL, POINT
            };
            status_t status = JUSTBEGAN;

            /* if end in JUSTBEGAN or NEWEVALUATE, raise red flag
             *  graph of possible status transitions
             *                             ____________
             *                            /            \
             *                            |            |
             *                            V            |
             * JUSTBEGAN ---------> NEWEVALUATE ---> POINT
             *                      |  ^               ^
             *                      |  |               |
             *                      V  |               |
             *                      GLOBAL-------------/
             */
            enum global_status_t {
                GLOBEGAN, TOREADPHAN, TOREADGLOB
            };
            // this is assigned based on the string parsed
            global_status_t globalstatus = GLOBEGAN;
            bool ADD_POINTMEAS = false;
            typedef char* inputline_t;
            inputline_t myline;
            measline_t mymeasline;
            execname_t myexecname;
            bool ignoredonce = false;
            myline = (&(line[0])); // myline permanently points to line[0]
            mymeasline = (&(line[0]));
            myexecname = (&(line[0]));
            while (fgets(line, maxlinelength, fp)) {
                ////	if(isspace(mymeasline)) continue;
                char *p;
                p = myline;
                while ((' ' == *p) || ('\t' == *p)) p++;
                if (('\n' == *p) || ('\r' == *p) || ('\0' == *p)) continue;
                switch (status) {
                    case JUSTBEGAN:
                        // first line does not have to start with evaluation name: allow comments
                        if (strstr(myline, "evaluation name:")) {
                            status = NEWEVALUATE;
                            try {
                                myEES.new_exec(myexecname); // READ IN evaluation name!
                            } catch (std::exception &e) {
                                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                            }
                        } else {
                            if (!ignoredonce) {
                                std::cout << "stuff in eval file before first 'evaluation name:' ignored!!!" << std::endl;
                                ignoredonce = true;
                            }
                        }
                        break; // the switch, not the while
                    case NEWEVALUATE:
                        if (strstr(myline, "global resolution measures")) {
                            status = GLOBAL;
                        } else {
                            if (strstr(myline, "point by point resolution measures")) {
                                status = POINT;
                            }
                        }
                        break; // out of switch
                    case GLOBAL:
                        if (strstr(myline, "evaluation name:")) { // can happen at any time
                            globalstatus = GLOBEGAN;
                            status = NEWEVALUATE;
                            try {
                                myEES.new_exec(myexecname); // READ IN evaluation name!
                            } catch (std::exception &e) {
                                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                            }
                        } else {
                            switch (globalstatus) {
                                case GLOBEGAN:
                                    if (strstr(myline, "hantom name")) {
                                        // read the name of phantom! (necessary?)
                                    } else if (strstr(myline, "etrics for test phantom")) {
                                        globalstatus = TOREADPHAN;
                                    }
                                    break; // out of inner switch
                                case TOREADPHAN: // read exactly one line of phantom data
                                    if (sscanf(myline, "%d", &tmpint) == 1) { // if line starts with an integer
                                        try {
                                            myEES.set_phantomeas(mymeasline);
                                        } catch (std::exception &e) {
                                            std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                                        }
                                        globalstatus = TOREADGLOB; // next, should try to read an execution
                                    }
                                    break; // out of inner switch
                                case TOREADGLOB:
                                    if (sscanf(myline, "%d", &tmpint) == 1) {
                                        try {
                                            myEES.add_to_globmeaset(mymeasline);
                                        } catch (std::exception &e) {
                                            std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                                        }
                                    } else if (strstr(myline, "execution name")) {
                                        // parse in execution name (necessary?)
                                        try {
                                            myEES.new_globmeaset(myexecname);
                                        } catch (std::exception &e) {
                                            std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                                        }
                                    } else if (strstr(myline, "etrics for algorithm")) {
                                        // parse in ALGO name! (maybe?)
                                    } else if (strstr(myline, "point by point resolution measures")) {
                                        globalstatus = GLOBEGAN;
                                        status = POINT;
                                    }
                                    break; // out of inner switch (optional here, defensive)
                            }
                        } // else
                        break; // out of switch
                    case POINT:
                        if (strstr(myline, "evaluation name")) {
                            // parse evaluation name!
                            ADD_POINTMEAS = false;
                            status = NEWEVALUATE;
                            try {
                                myEES.new_exec(myexecname); // READ IN evaluation name!
                            } catch (std::exception &e) {
                                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                            }
                        } else if (strstr(myline, "execution name")) {
                            ADD_POINTMEAS = false;
                            try {
                                myEES.new_PBPmeaset(myexecname); // name gets assigned as we go along, but still parse in name
                            } catch (std::exception &e) {
                                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                            }
                        } else if (strstr(myline, "algn")) { // need a line with this
                            ADD_POINTMEAS = true;
                        } else if (ADD_POINTMEAS) {
                            // NOW we can go ahead with this, since whitespace-lines are skipped
                            try {
                                myEES.add_to_PBPmeaset(mymeasline);
                            } catch (std::exception &e) {
                                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                            }
                        }
                        break; // out of switch (optional here, defensive)
                } // status switch
                memset(myline, 0, maxlinelength); // reinit line to null terminator
            } // while can keep reading lines
            if ((status == JUSTBEGAN) || (status == NEWEVALUATE)) {
                //maybe	throw std::runtime_error("malformed eval file: no measurements");
                std::cerr << "malformed eval file: no measurements... please open a valid eval file." << std::endl;
                myEES.clear();
                return;
            }
            try {
                myEES.complete();
            } catch (std::exception &e) {
                std::cerr << "This badness happened:" << std::endl << e.what() << std::endl;
                myEES.clear();
                return;
            }
            /// Call Eval Window
            //   new version: first, user should select executions!
            QPoint topleft;
            topleft = this->pos();
            selectEvalExecutionsWindow* selectEvalExecsWin;
            selectEvalExecsWin = new selectEvalExecutionsWindow(desk, "Open Evaluate Executions", false, 0, topleft.x() + 30, topleft.y() + 30);
            selectEvalExecsWin->show();
        } // --if was able to open the file ("else" of failure, failure already handled)
    } // --if QDialog accepted
    // else?
} // --SnarkDisplay::openEvalFile()

/** Updates low and high thresholds for image windows having the button Global Thresholds checked
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::updateGlobalThreshs() {
    // the new way: do it for all of them!
    for (int i = 0; i < numimages; i++)
        if (imageset[i]->isLoaded() && displaywindowset[i]) {
            displaywindowset[i]->setGlobalthreshs();
        }
}

/** tells all windows to be activated by the global thresholds
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::signalAllGlobals(bool on) {
    allglobal = on;
    if (myRangeWinP) {
        myRangeWinP->signalAllGlobals(allglobal);
    }
    for (int i = 0; i < numimages; i++) {
        if (alldifference && (i == 0)) continue;
        if (imageset[i]->isLoaded() && displaywindowset[i]) {
            displaywindowset[i]->setGlobal(on);
        }
    }
    updateGlobalThreshs(); // refresh immediately
}

void SnarkDisplay::signalAllDifference(bool on) {
    alldifference = on;
    if (myRangeWinP) {
        myRangeWinP->signalAllDifference(alldifference);
    }
    if (allglobal) {
        if (imageset[0]->isLoaded() && displaywindowset[0]) {
            if (alldifference) {
                displaywindowset[0]->setGlobal(false);
            } else {
                displaywindowset[0]->setGlobal(true);
            }
        }
    }
    for (int i = 1; i < numimages; i++) {
        if (imageset[i]->isLoaded() && displaywindowset[i]) {
            displaywindowset[i]->setDifference(on);
        }
    }
}

void SnarkDisplay::signalAllIntensity(bool on) {
    for (int i = 0; i < numimages; i++) {
        if (displaywindowset[i]) {
            displaywindowset[i]->setIntensity(on);
        }
    }
}

/** Calls about window
@param void
@author Bruno M. Carvalho & Deniz Sarioz
@version 2.0 */
void SnarkDisplay::about() {
    QMessageBox::information(this, "snark14Display",
            "snark14Display was developed by\n"
            "Bruno M. Carvalho and Deniz Sarioz\n"
            "2003--2008\n"
            "copyright (QPL) by CUNY-GC-CS-DIG\n"
            "http://www.dig.cs.gc.cuny.edu");
}

/** Allocates space for a vector of Snarkimages
@param ni number of images
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::allocSnarkimages(int ni) {
    int i;
    imageset = (Snarkimage **) malloc((unsigned) (ni) * sizeof (Snarkimage*));
    if (!imageset) {
        QMessageBox::information(this, "SnarkDisplay", "Error!\n"
                "Allocation failure in allocSnarkimages()\n");
    } else {
        for (i = 0; i < ni; i++) {
            imageset[i] = new Snarkimage();
            if (!imageset[i]) {
                QMessageBox::information(this, "SnarkDisplay", "Error!\n"
                        "Allocation failure in allocSnarkimages()\n");
            }
        }
    }
}

/** Allocates space for a vector of displaywindow POINTERS
@param ni number of images
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::allocDisplaywindows(int ni) {
    allglobal = false;
    alldifference = false;
    globalmax = globaldiffmax = -100000;
    globalmin = globaldiffmin = 100000;
    // forget everything we did before!
    if (ni == 0) return;
    //  displaywindowset = (displaywindow **) malloc((unsigned) (ni)*sizeof(displaywindow*));
    displaywindowset = new displaywindow* [ni];
    if (!displaywindowset) {
        QMessageBox::information(this, "SnarkDisplay", "Error!\n"
                "Allocation failure in allocDisplaywindows()\n");
    }
    displaywindowactive = (bool *) malloc((unsigned) (ni) * sizeof (bool));
    for (int i = 0; i < ni; i++) {
        displaywindowactive[i] = false;
        displaywindowset[i] = 0; // NULL
    }
}

/** Frees space of a vector of Snarkimages
@param ni number of images
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::freeSnarkimages(int ni) {
    int i;

    if (imageset) {
        for (i = 0; i < ni; i++) {
            if (imageset[i]) {
                imageset[i]->~Snarkimage();
                imageset[i] = 0;
            }
        }
        free(imageset);
    }
} // --SnarkDisplay::freeSnarkimages()

/** Frees space of a vector of Displaywindows
@param void
@author Bruno M. Carvalho
@version 1.0 */
void SnarkDisplay::freeDisplaywindows(int ni) {
    int i;
    if (displaywindowset) {
        for (i = 0; i < ni; i++) {
            if (displaywindowset[i]) {
                delete displaywindowset[i];
            }
        }
        delete [] displaywindowset;
    }
} // --SnarkDisplay::freeDisplaywindows()

/** Starts the application
@param void
@author Bruno M. Carvalho, parametrized by deniz
@version 1.2 */
int main(int argc, char* argv[]) {
    QApplication snarkapp(argc, argv);

    if (argc == 3) {
        int wid = atoi(argv[1]);
        int hei = atoi(argv[2]);
        int minwid = 3 * sd_line_t::XWINPAD;
        int minhei = 3 * sd_line_t::YWINPAD;
        if ((wid >= minwid) && (hei >= minhei)) {
            std::cout << "For plots, accepted window WIDTH=" << wid << ", HEIGHT=" << hei << std::endl;
            sd_line_t::WIDTH = wid;
            sd_line_t::HEIGHT = hei;
        } else {
            std::cout << "Minimum width: " << minwid << ", minimum height: " << minhei << std::endl;
            std::cout << "Aborting " << argv[0] << std::endl;
            return 1;
        }
    }
    //  SnarkDisplay* mw=new SnarkDisplay();
    SnarkDisplay mw;
    //  snarkapp.setMainWidget(mw);
    snarkapp.setMainWidget(&mw);
    //  mw->show();
    mw.show();

    return snarkapp.exec();
} // --main()
