/** @file eval_helper.hpp
    @package snark14Display
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
*/

/** @name eval_helper.hpp
    @type C++ header file
    @date April 2004
    @package snark14Display
    @see eval_helper.cpp
*/

#ifndef _EVAL_HELPER_HPP 
#define _EVAL_HELPER_HPP

#include <qwidget.h>
#include "verbosity.hpp"
#include <exception>
#include <stdexcept>
#include <cstdio>
#include <iostream>
#include <deque>
#include <vector>

#include "line_real_set_t.hpp"


enum globalYVar_t { AVERAGE, DISTANCE, RELERR, VARIANCE, STDDEV, RESIDUAL, KULLBACK };

typedef char* measline_t; // these types are potentially used everywhere
typedef char* execname_t;

class phantomeas;        // forward declaration of classes used by eval_execution_set
class eval_execution;

/** "public interface"-like class definition of eval_execution_set: 
 *    superstructure capable of holding information 
 *    from many eval executions, an entire eval08 file  */
class eval_execution_set {
  std::deque<eval_execution> execs; // entire output of snark14 EVALUATE executions
  std::vector<bool> selected; // holds which execs are selected by snark14Display user
  phantomeas* phantomeasp;
  bool completed; // holds whether creation of execs is complete (is this really necessary here?)
  bool extrema_calculated; // holds whether extrema are calculated (ditto.)
  bool selectionHasResidualvar;
  bool selectionHasKullbackvar;
  bool same_x_forall;  // for iter, plotted on x-axis
  bool same_y_forall;  // for whatever it is that is requested for plotting on y-axis
  // (almost the same as for eval_execution) extrama...
  // first the global extrema
  int minIter, maxIter;
  double minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr;
  double minVariance, maxVariance, minStddev, maxStddev, minResidual, maxResidual, minKullback, maxKullback;
  // then the PBP extrema...
  std::vector<double> PBP_minima; // minima and maxima of each E(0)..E(N-1), N not varying within an eval file.
  std::vector<double> PBP_maxima; // to be calculated over ALL EVALUATE calls, depending.
  unsigned num_Es; // 'N' in comment above
  // e.g., PBP_minima[2] is the minimum value of all E(2) in all PBPmeasets in all execs.
public:
  // construction
  eval_execution_set();
  // cultivation  (call during the eval file reading phase)
  void new_exec(execname_t line);
  void new_globmeaset(execname_t line);
  void add_to_globmeaset(measline_t line);
  void new_PBPmeaset(execname_t line);
  void add_to_PBPmeaset(measline_t line);
  void set_phantomeas(measline_t line);
  // completion (call to signal transition from cultivation phase to manipulation phase)
  void complete();
  // manipulation
  bool calculate_global_extrema(bool restrict_domain, int min_it_arg, int max_it_arg);
  bool calculate_global_extrema();
  bool calculate_PBP_extrema(bool restrict_domain, int min_it_arg, int max_it_arg);
  bool calculate_PBP_extrema();
  void deselect_exec(unsigned idx);
  void select_exec(unsigned idx);
  void deselect_all_execs();
  void select_all_execs();
  void setSameXForAll (bool val = true);
  void setSameYForAll (bool val = true);
  void useSameExtremaForAll (bool val = true);
  void useDifferentExtremaForEach();
  // information extraction pertaining to entire set
  unsigned num_eval_execs();
  bool hasPhantom();
  bool selectionHasGlobal();
  bool selectionHasResidual(); // should really be called only after selectionHasGlobal()
  bool selectionHasKullback();
  bool selectionHasPBP();
  unsigned get_num_Es(); // if no PBP, returns 0.  should really be called only after hasPBP()
  bool getSameXForAll();
  bool getSameYForAll();
  int getMinIterViaGlobal();   // by some sleight of theory, 
  int getMaxIterViaGlobal();  // iterations could be different from global to PBP... 
  int getMinIterViaPBP();    // I'm taking out insurance.
  int getMaxIterViaPBP();
  double getMinYGlobal(globalYVar_t globalYVar);
  double getMaxYGlobal(globalYVar_t globalYVar);
  double getMinYPBP(unsigned idx);
  double getMaxYPBP(unsigned idx);
  // information extraction pertaining to a single evaluate execution
  execname_t getExecName(unsigned idx);
  // reporting to stdout (for debugging)
  void show_all();
  void show_selected();
  void show_extrema();
  void addPBPVarLines(line_real_set_t& LRS, unsigned PBPYVar, unsigned ee_index);
  void addGlobVarLines(line_real_set_t& LRS, globalYVar_t globalYVar, unsigned ee_index);
  void spawnDialogsGlob(QWidget* parent, globalYVar_t globalYVar);
  void spawnDialogsPBP(QWidget* parent, unsigned PBPYVar);
  void spawnAllPlotsGlob(QWidget* parent, globalYVar_t globalYVar, int userMinXVal, int userMaxXVal, double userMinYVal, double userMaxYVal, bool isGrayScale);
  void spawnAllPlotsPBP(QWidget* parent, unsigned PBPYVar, int userMinXVal, int userMaxXVal, double userMinYVal, double userMaxYVal, bool isGrayScale);
  // two flavors of destruction
  void clear(); // allows there to be a single, shared variable
  ~eval_execution_set();
}; // --"public interface" (within entire snark14Display) of class eval_execution_set
#endif // _EVAL_HELPER_HPP
// --eval_helper.hpp
