/** @file plot_t.cpp
    @author Deniz
    @description functions in class plot_t is implemented here
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
 */
#include <qcolor.h>
#include <cmath>
#include <list>
#include <exception>
#include <stdexcept>
#include <iostream>
#include <qpainter.h>
#include <qstring.h>

#include "verbosity.hpp"
#include "plot_t.hpp"
#include "line_window_onevar_t.hpp"
#include "line_real_set_t.hpp"
#include "line_real_t.hpp"

/** ultimate window-coord class: holds SEVERAL vars' info and can plot them in different colors */


typedef std::list<line_window_onevar_t> data_manyexecs_t;

plot_t::plot_t(line_real_set_t& LRS, int WIDTH, int HEIGHT) {
    // can also take padding arguments maybe?
    // we are a friend of line_real_set_t
    plotName = "garden variety plot";
    XAxisName = "x-units";
    YAxisName = "y-units";
    rMinX = LRS.getMinX();
    rMaxX = LRS.getMaxX();
    rMinY = LRS.getMinY();
    rMaxY = LRS.getMaxY();
    // calculate instance variables, mainly to make the right label notches on axes
    sd_line_t::WIDTH = WIDTH;
    (this->myWidth) = WIDTH;
    sd_line_t::HEIGHT = HEIGHT;
    (this->myHeight) = HEIGHT; // insurance
    // perform fudging defensively
    if (0 < sd_line_t::X_HALF_FUDGE_INCREMENT) {
        rMinX -= sd_line_t::X_HALF_FUDGE_INCREMENT;
        rMaxX += sd_line_t::X_HALF_FUDGE_INCREMENT;
    }
    xr_range = rMaxX - rMinX; // underscore between r's for emphasis
    yr_range = rMaxY - rMinY;
    if (0 < sd_line_t::Y_HALF_FUDGE_FACTOR) {
        double y_half_fudge = yr_range * sd_line_t::Y_HALF_FUDGE_FACTOR;
        rMinY -= y_half_fudge;
        rMaxY += y_half_fudge;
        yr_range = rMaxY - rMinY; // calculate post-fudge yr_range
    }
    ywmin = sd_line_t::YWINPAD;
    ywmax = (this->myHeight) - 1 - sd_line_t::YWINPAD;
    ywrange = ywmax - ywmin;
    xwmin = sd_line_t::XWINPAD;
    xwrange = (this->myWidth) - 1 - 2 * sd_line_t::XWINPAD;
    ywmin = sd_line_t::YWINPAD;
    ywmax = (this->myHeight) - 1 - sd_line_t::YWINPAD;
    xscalar = static_cast<double> (xwrange) / static_cast<double> (xr_range);
    yscalar = static_cast<double> (ywrange) / yr_range;
    data.clear(); // JUST in case
    // LRS has a bunch of line_real_t's inside
    line_real_set_t::line_real_list_t::const_iterator it = LRS.data.begin();
    for (; it != LRS.data.end(); it++) {
        line_real_t LR = (*it);
        line_window_onevar_t LWO(LR);
        if (std::verbose > 2) {
            LWO.show();
        }
        data.push_back(LWO);
    }
} // --plot_t ctor/3

/** give a name to this plot_t object */
void plot_t::setName(QString& inName) {
    plotName = inName;
} // --plot_t::setName()

void plot_t::setXAxisName(QString& inNameX) {
    XAxisName = inNameX;
} // --plot_t::setXAxisName()

void plot_t::setYAxisName(QString& inNameY) {
    YAxisName = inNameY;
} // --plot_t::setYAxisName()

/** calculate and make appropriate and human-oriented range increment for Y Axis:
 *  1, 2, or 5 times a power of 10
 *  such there are usually about 10 to 20 marks on the axis
 *  this calibration can be fudged to about halve or double the number of notches
 */
void plot_t::makeNotchesOnYAxis(QPainter& P) {
    double rmin = this->rMinY; // minimum as a real
    double rmax = this->rMaxY; // maximum as a real
    double range = rmax - rmin; // range: real
    int xPosOfYAxis = xwmin;
    // this code is for calculating the values at which we make notches on the axes
    double logbase10ofrange = log(range) / log(10);
    int characteristic = static_cast<int> (floor(logbase10ofrange));
    double mantissa = logbase10ofrange - characteristic;
    double log2 = log(2) / log(10);
    double log5 = log(5) / log(10);
    int multiplierTimes10;
    /* the whole idea is to partition numbers like a partition of frequencies \in |R+
     *     by musical note-based half-open intervals, e.g.: { [A,C#) , [C#,F) , [F,A) }
     * we partition all ranges \in |R+ into three representatives: {1,2,5}.
     * so we work on the mantissa of the range's log base 10, by definition \in [0,1).
     * unlike the base 12-partitioned base 2 musical octaves, we throw away and add back in powers of 10,
     *     which the snark14Display user tribe feels comfortable with.
     *          ( we really tripartition each interval [r, 10r) where r \in |R+ )
     * since our representatives are {1, 2, 5}.10^n (does not form a perfect geometric sequence when sorted),
     *     we partition [0,1) into { [0,log2) , [log2,log5) , [log5, 1) } => represented as {1,2,5}
     * rather than using the perfectly even tri-partition of notes shown above.
     * say range is 763, then the mantissa will be log(7.63), which falls in the third partition,
     * so all intervals are 50 (5 times the appropriate power of 10 remembered by characteristic)
     * and range/notchinterval = 763/50 is indeed in [10,20).  get either 15 or 16 notches, depending on extrema.
     * in the case of something like range = 4.9, we have 0.2 be the representative: 24 or 25 notches.
     *   (that is the absolute maximum of the number of notches.)
     */
    if (mantissa < log2) {
        multiplierTimes10 = 1;
    } else if (mantissa < log5) {
        multiplierTimes10 = 2;
    } else {
        multiplierTimes10 = 5;
    }
    //  double range_inc = multiplier * pow(10, characteristic);
    double range_inc = multiplierTimes10 * pow(10, characteristic - 1);
    //    double from = range_inc * ceil(rmin/range_inc);
    // handle "weird zero" (bug #114) this way...
    int from_idx = static_cast<int> (ceil(rmin / range_inc));
    // not idiomatic 'C' but it is easier for me to think about inclusive bounds
    //    double to   = range_inc * floor(rmax/range_inc);
    int to_idx = static_cast<int> (floor(rmax / range_inc));
    //    int num_intervals = static_cast<int>(floor((to-from)/range_inc));
    P.setPen(Qt::black);
    QFont f("times", 12);
    P.setFont(f);
    //    for(int notch = 0; notch <= num_intervals; notch++) {
    for (int idx = from_idx; idx <= to_idx; idx++) {
        double markreal = idx * range_inc; // if idx==0, markreal will be 0.0
        int markwinY = static_cast<int> (floor(ywmax - (markreal - rMinY) * yscalar)); // mark position in window
        P.drawLine(xPosOfYAxis - 6, markwinY, xPosOfYAxis + 1, markwinY);
        QString labelToDraw;
        labelToDraw.setNum(markreal);
        P.drawText(xPosOfYAxis - sd_line_t::YAXISNOTCHLABELOFFSET, markwinY, labelToDraw);
    }
} // --plot_t::makeNotchesOnYAxis()

/** just like the above, slightly trickier since we only want integers
 */
void plot_t::makeNotchesOnXAxis(QPainter& P) {
    int& imin = (this->rMinX);
    int& imax = (this->rMaxX);
    int yPosOfXAxis = ywmax;
    double rmin = static_cast<double> (imin); // minimum as a real
    double rmax = static_cast<double> (imax); // maximum as a real
    double range = rmax - rmin; // range: real
    // this code is for calculating the values at which we make notches on the axes
    double logbase10ofrange = log(range) / log(10);
    int characteristic = static_cast<int> (floor(logbase10ofrange));
    double mantissa = logbase10ofrange - characteristic;
    double log2 = log(2) / log(10);
    double log5 = log(5) / log(10);
    int multiplierTimes10; // i like using integral types.
    // the logic for this whole operation is the same as above,
    // save some measures taken to ensure that the interval as well as notches are always integers.
    if (mantissa < log2) {
        multiplierTimes10 = 1;
    } else if (mantissa < log5) {
        multiplierTimes10 = 2;
    } else {
        multiplierTimes10 = 5;
    }
    // the following line really means: // double range_inc = multiplier * pow(10, characteristic);
    double range_inc = multiplierTimes10 * pow(10, characteristic - 1);
    if (range_inc <= 1) range_inc = 1.0; // x-axis should not have fractional labels (>1 guarantees it isn't)
    int range_inc_int = static_cast<int> (floor(range_inc));
    if (range_inc_int <= 0) range_inc_int = 1; // paranoia: don't want loop increment to be 0
    int from_int = static_cast<int> (floor(range_inc * ceil(rmin / range_inc)));
    // not idiomatic 'C' but it is easier for me to think about inclusive bounds
    int to = static_cast<int> (floor(range_inc * floor(rmax / range_inc)));
    P.setPen(Qt::black); // darkish blue notches and axis labels [as was in Bruno's code]
    QFont f("times", 12);
    P.setFont(f);
    for (int notch = from_int; notch <= to; notch += range_inc_int) {
        // safe to increment by integer in for loop
        int markreal = notch; // really!
        int markwinX = xwmin + static_cast<int> (floor((markreal - rMinX) * xscalar));
        P.drawLine(markwinX, yPosOfXAxis - 1, markwinX, yPosOfXAxis + 6);
        QString labelToDraw;
        labelToDraw.setNum(markreal);
        P.drawText(markwinX - 5, yPosOfXAxis + 19, labelToDraw);
    }
} // --plot_t::makeNotchesOnXAxis()

//    void plot_t::labelTheXAxis() {
//    } // --plot_t::labelTheXAxis()
//    void plot_t::labelTheYAxis() {
//    } // --plot_t::labelTheYAxis()

/** plot the plot_t object */
void plot_t::plot(QPainter& P) {
  static const int numCanned = 15;
  static QColor cannedColors[numCanned] = {
    Qt::red,          //  0
    Qt::green,
    Qt::blue,
    Qt::cyan,
    Qt::magenta,
    Qt::yellow,       //  5
    Qt::gray,
    Qt::darkRed,
    Qt::darkGreen,
    Qt::darkBlue,
    Qt::darkCyan,     // 10
    Qt::darkMagenta,
    Qt::darkYellow,
    Qt::black,
    Qt::lightGray     // 14
  };
  static const int numPenStyles = 5;
  static Qt::PenStyle  cannedPenStyles[numPenStyles] = {
    Qt::SolidLine,
    Qt::SolidLine,
    Qt::SolidLine,
    Qt::SolidLine,
    Qt::SolidLine
  };
  int RectWidth = 16;  // swatch rectangles' (for algo color/name association) dimensions
  int RectHeight = 16;
  int xPadForRectCtrs = RectWidth;
  int yPadForRectCtrs = RectHeight;
  int xCtrDistance = myWidth/3;
  int yCtrDistance = ywmin/4; // so it has to be that ywmin > 4*RectHeight
  // for drawing legend-swatches: 2x3 on top, 3x3 at bottom, total of 15
  static point_window_t RectCtrPositions[numCanned] = {
    // swatches above the plot
    {xPadForRectCtrs + 0 * xCtrDistance,  yPadForRectCtrs + 0 * yCtrDistance},
    {xPadForRectCtrs + 1 * xCtrDistance,  yPadForRectCtrs + 0 * yCtrDistance},
    {xPadForRectCtrs + 2 * xCtrDistance,  yPadForRectCtrs + 0 * yCtrDistance},
    {xPadForRectCtrs + 0 * xCtrDistance,  yPadForRectCtrs + 1 * yCtrDistance},
    {xPadForRectCtrs + 1 * xCtrDistance,  yPadForRectCtrs + 1 * yCtrDistance},
    {xPadForRectCtrs + 2 * xCtrDistance,  yPadForRectCtrs + 1 * yCtrDistance},
    // leave room above the plot for plot name... and swatches below the plot
    {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance},
    {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance},
    {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance},
    {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
    {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
    {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
    // really shouldn't plot this many... these could clash a bit with x axis label.
    {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance},
    {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance},
    {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance}
  };
  int colorInd = 0; // don't want this to be static!
  int penStyleInd = 0;
  for(data_manyexecs_t::iterator it = data.begin(); it != data.end(); it++) {
    QColor& myColor = cannedColors[ (colorInd % numCanned) ];
    Qt::PenStyle myPenStyle = cannedPenStyles[ penStyleInd % numPenStyles ];
    QPen pen = P.pen();
    //pen.setStyle(myPenStyle);
    pen.setWidth(5);
    P.setPen(pen);

    it->plot(P, myColor, myPenStyle);
    if(colorInd < numCanned) {
      int RectCtrX = RectCtrPositions[colorInd].x;
      int RectCtrY = RectCtrPositions[colorInd].y;
      // fill in rectangle with corresponding color (same as the line just plotted)
      P.fillRect(RectCtrX - RectWidth/2, RectCtrY - RectHeight/2, RectWidth, RectHeight, myColor);
      P.setPen(Qt::black); // then draw the boundary
      P.drawRect(RectCtrX - RectWidth/2, RectCtrY - RectHeight/2, RectWidth, RectHeight);
      QString algName = (it->getName());
      int nameLengthInChars = 40; // max
      QFont savedFont = P.font();
      QFont f("Times", 11);
      P.setFont(f);
      P.drawText(RectCtrX + RectWidth*3/4, RectCtrY+5, algName, nameLengthInChars);
      P.setFont(savedFont);
    } else {
      // perhaps generate some kind of warning about too many plots,
      // not enough colors being used / not enough space for swatches etc
    }
    // no one in their right mind would run 16 different execs (?), but just covering our base (no segfault).
    // here could also be where label associating color with variable name is plotted...
    colorInd++;
    penStyleInd++;
  }
  sd_line_t::drawBoundary(P); // by doing this after, we make sure the frame is not painted over
  makeNotchesOnXAxis(P);
  makeNotchesOnYAxis(P);
  // now label the plot just above the plot window
  QFont F = P.font();
  QFont FBold = F;
  FBold.setBold(true);
  P.setFont(FBold);
  P.setPen(Qt::black);
  P.drawText(xwmin, ywmin-15, plotName);
  P.setFont(F); // restore original font to QPainter
  // label the X axis
  P.drawText( (sd_line_t::WIDTH - sd_line_t::XWINPAD)/2, ywmax+45, XAxisName);
  // label the Y axis (sideways)
  P.rotate(-90);
  P.drawText( -(sd_line_t::HEIGHT)/2, sd_line_t::XWINPAD - sd_line_t::YAXISLABELOFFSET, YAxisName);
  P.rotate(90); // restore orientation of painter
} // --plot_t::plot()

/** plot the plot_t object in Grayscale*/
// jk 1/13/2009 adding grayScale option for graphs
void plot_t::plotGray(QPainter& P) {

    static const int numCanned = 9;
    static QColor cannedColors[numCanned] = {
        Qt::gray,
        Qt::black,
        Qt::lightGray,
        Qt::gray,
        Qt::black,
        Qt::lightGray,
        Qt::gray,
        Qt::black,
        Qt::lightGray,
    };
    static const int numPenStyles = 9;
    static Qt::PenStyle cannedPenStyles[numPenStyles] = {
        Qt::SolidLine,
        Qt::SolidLine,
        Qt::SolidLine,
        Qt::DotLine,
        Qt::DotLine,
        Qt::DotLine,
        Qt::DashLine,
        Qt::DashLine,
        Qt::DashLine
        //Qt::DashDotLine,
        //Qt::DashDotDotLine,
    };
    int RectWidth = 16; // swatch rectangles' (for algo color/name association) dimensions
    int RectHeight = 16;
    int xPadForRectCtrs = RectWidth;
    int yPadForRectCtrs = RectHeight;
    int xCtrDistance = myWidth / 3;
    int yCtrDistance = ywmin / 4; // so it has to be that ywmin > 4*RectHeight
    // for drawing legend-swatches: 2x3 on top, 3x3 at bottom, total of 15
    static point_window_t RectCtrPositions[numCanned] = {
        // swatches above the plot
        {xPadForRectCtrs + 0 * xCtrDistance, yPadForRectCtrs + 0 * yCtrDistance},
        {xPadForRectCtrs + 1 * xCtrDistance, yPadForRectCtrs + 0 * yCtrDistance},
        {xPadForRectCtrs + 2 * xCtrDistance, yPadForRectCtrs + 0 * yCtrDistance},
        {xPadForRectCtrs + 0 * xCtrDistance, yPadForRectCtrs + 1 * yCtrDistance},
        {xPadForRectCtrs + 1 * xCtrDistance, yPadForRectCtrs + 1 * yCtrDistance},
        {xPadForRectCtrs + 2 * xCtrDistance, yPadForRectCtrs + 1 * yCtrDistance},
        // leave room above the plot for plot name... and swatches below the plot
        {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance},
        {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance},
        {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 0 * yCtrDistance}
        //    {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
        //    {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
        //    {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 1 * yCtrDistance},
        //    // really shouldn't plot this many... these could clash a bit with x axis label.
        //    {xPadForRectCtrs + 0 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance},
        //    {xPadForRectCtrs + 1 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance},
        //    {xPadForRectCtrs + 2 * xCtrDistance, myHeight - yPadForRectCtrs - 2 * yCtrDistance}
    };
    int colorInd = 0; // don't want this to be static!
    int penStyleInd = 0;
    for (data_manyexecs_t::iterator it = data.begin(); it != data.end(); it++) {
        QColor& myColor = cannedColors[ (colorInd % numCanned) ];
        Qt::PenStyle myPenStyle = cannedPenStyles[ penStyleInd % numPenStyles ];

        it->plot(P, myColor, myPenStyle);
        if (colorInd < numCanned) {
            int RectCtrX = RectCtrPositions[colorInd].x;
            int RectCtrY = RectCtrPositions[colorInd].y;
            // fill in rectangle with corresponding color (same as the line just plotted)
            //P.fillRect(RectCtrX - RectWidth/2, RectCtrY - RectHeight/2, RectWidth, RectHeight, myColor);
            QPen savedPen = P.pen();
            QPen pen; // = P.pen();
            pen.setStyle(myPenStyle);
            pen.setWidth(2);
            pen.setColor(myColor);
            P.setPen(pen);

            //P.setPen(myColor); // then draw the boundary
            //P.setPen(myPenStyle);
            //P.drawRect(RectCtrX - RectWidth/2, RectCtrY - RectHeight/2, RectWidth, RectHeight);
            P.drawLine(RectCtrX - RectWidth / 2, RectCtrY, RectCtrX + RectWidth / 2, RectCtrY);
            //change pen color and style back to black solid for printing the text
            //P.setPen( savedPen );
            pen.setStyle(Qt::SolidLine);
            pen.setWidth(1);
            pen.setColor(Qt::black);
            P.setPen(pen);

            QString algName = (it->getName());
            int nameLengthInChars = 40; // max
            QFont savedFont = P.font();
            QFont f("Times", 11);
            P.setFont(f);
            P.drawText(RectCtrX + RectWidth * 3 / 4, RectCtrY + 5, algName, nameLengthInChars);
            P.setFont(savedFont);

        } else {
            // perhaps generate some kind of warning about too many plots,
            // not enough colors being used / not enough space for swatches etc
        }
        // no one in their right mind would run 16 different execs (?), but just covering our base (no segfault).
        // here could also be where label associating color with variable name is plotted...
        colorInd++;
        penStyleInd++;
    }
    sd_line_t::drawBoundary(P); // by doing this after, we make sure the frame is not painted over
    makeNotchesOnXAxis(P);
    makeNotchesOnYAxis(P);
    // now label the plot just above the plot window
    QFont F = P.font();
    QFont FBold = F;
    FBold.setBold(true);
    P.setFont(FBold);
    P.setPen(Qt::black);
    P.drawText(xwmin, ywmin - 15, plotName);
    P.setFont(F); // restore original font to QPainter
    // label the X axis
    P.drawText((sd_line_t::WIDTH - sd_line_t::XWINPAD) / 2, ywmax + 45, XAxisName);
    // label the Y axis (sideways)
    P.rotate(-90);
    P.drawText(-(sd_line_t::HEIGHT) / 2, sd_line_t::XWINPAD - sd_line_t::YAXISLABELOFFSET, YAxisName);
    P.rotate(90); // restore orientation of painter
} // --plot_t::plot()

